"""
isicarchive.font (Font)

This module provides the Font helper class and doesn't have to be
imported from outside the main package functionality (IsicApi).

To instantiate a Font object, simply call it with the font name:

>>> font = isicarchive.font.Font('calibri')

At this time, only calibri is available as a font file!

The format is similar to that used by
https://github.com/neuroelf/neuroelf-matlab/
in the image_font function of the @neuroelf class. It uses an
image with all available extended ASCII characters, and computes
a kerning value between all combinations of letters for the type
setting.

To then create an image with a set text, you can call either the
```font.set_line(TEXT)``` method for a one-channel uint8 image,
or the ```font.set_text(TEXT, ...)``` method for an RGB and alpha
image with additional options.
"""

# in large parts, code was translated from
# https://github.com/neuroelf/neuroelf-matlab/blob/master/%40neuroelf/private/image_font.m


# specific version for file
__version__ = '0.4.8'


# imports (needed for majority of functions)
import os
from typing import Tuple, Union

from imageio import imread
import numpy


class Font(object):
    """
    Font
    """

    __fonts = {}
    def __init__(self, fontname:str):

        # setup object
        self._image = None
        self._kerning = None
        self._letters = None
        self._lmap = 0 - numpy.ones(1024, dtype=numpy.int32)
        self._num_letters = 0
        self._size = 0
        self._xktab = None
        self._xstart = None
        self._xstop = None
        self._ybase = 0
        self._yret = None
        self.name = ''

        # parse input (name = filename)
        if fontname is None or fontname == '':
            fontname = 'calibri'
        else:
            fontname = fontname.lower()
        fontfolder = os.path.dirname(__file__) + os.sep + 'etc' + os.sep
        if os.path.exists(fontfolder + 'font_' + fontname + '.npz'):
            self.name = fontname
        else:
            self.name = 'calibri'
        if self.name in self.__fonts:
            f = self.__fonts[self.name]
            self._image = f['image']
            self._kerning = f['kerning']
            self._letters = f['letters']
            self._lmap = f['lmap']
            self._num_letters = f['num_letters']
            self._size = f['size']
            self._xktab = f['xktab']
            self._xstart = f['xstart']
            self._xstop = f['xstop']
            self._ybase = f['ybase']
            self._yret = f['yret']
            return

        # load font file and set in object
        fontfile = fontfolder + 'font_' + self.name + '.npz'
        fontdata = numpy.load(fontfile)
        fontdict = {k:v for (k,v) in fontdata.items()}
        self._image = imread(fontdict['fimage'].tobytes())
        self._letters = fontdict['letters']
        self._num_letters = fontdict['flen']
        self._lmap[self._letters] = numpy.asarray(range(self._num_letters))
        self._size = fontdict['size']
        nl = self._num_letters
        self._xktab = numpy.zeros(nl * nl, dtype=numpy.float32).reshape((nl,nl,))
        self._xstart = numpy.concatenate((numpy.zeros(1, dtype=numpy.int32), 
            fontdict['xstop'][0:-1]))
        self._xstop = fontdict['xstop']
        self._ybase = fontdict['ybase']
        for (d0,d1,v) in zip(fontdict['xk0'], fontdict['xk1'], fontdict['xkv']):
            self._xktab[d0-1,d1-1] = v
        self._yret = self._size - self._image.shape[0]
        self._add_kerning()
        self.__fonts[self.name] = {
            'image': self._image,
            'kerning': self._kerning,
            'letters': self._letters,
            'lmap': self._lmap,
            'num_letters': self._num_letters,
            'size': self._size,
            'xktab': self._xktab,
            'xstart': self._xstart,
            'xstop': self._xstop,
            'ybase': self._ybase,
            'yret': self._yret,
        }
        
    def __repr__(self):
        return 'isicarchive.font.Font(\'' + self.name + '\')'

    # font kerning
    def _add_kerning(self):

        # for each letter
        nl = self._num_letters
        fh = self._image.shape[0]
        fhp = numpy.asarray(range(-1,2)).reshape((1,3,))
        fhi = numpy.asarray(range(0,fh)).reshape((fh,1,)) + fhp
        fhi[0,0] = 0
        fhi[-1,-1] = fh - 1
        fhi.shape = (3 * fh,)
        lsf = numpy.zeros(fh * nl, dtype=numpy.int32).reshape((fh,nl,))
        lsf.fill(-65536)
        rsf = numpy.copy(lsf)
        for lc in range(1, nl):

            # get the letter image (masked)
            lmi = (self._image[:, self._xstart[lc]:self._xstop[lc]] >= 128)
            lms = self._xstop[lc] - self._xstart[lc]
            lmi = numpy.any(lmi[fhi,:].reshape((fh,3,lms,)), axis=1).reshape((fh,lms,))
            lms -= 1

            # find the first pixel that is not background
            cpix = numpy.where(numpy.sum(lmi, axis=1) > 1)[0]
            for cc in range(cpix.size):
                cpc = cpix[cc]
                cpw = numpy.where(lmi[cpc,:])[0]
                lsf[cpc,lc] = cpw[0]
                rsf[cpc,lc] = lms - cpw[-1]

        # next compute the median for each pair
        ktab = numpy.zeros(nl * nl, dtype=numpy.float32).reshape((nl, nl,))
        ktab.fill(numpy.nan)
        for rc in range(1,nl):
            rsfc = rsf[:,rc]
            if numpy.all(rsfc == -65536):
                continue
            nrf = numpy.sum(rsfc > -65536)
            for lc in range(1,nl):
                rsflsf = rsfc + lsf[:,lc]
                if all(rsflsf <= -32768):
                    continue
                nlf = numpy.float(numpy.sum(lsf[:,lc] > -65536))
                rsflsf = rsflsf[rsflsf > -32768]
                nrl = numpy.float(rsflsf.size)
                rlmin = numpy.float(numpy.amin(rsflsf))
                rlmed = numpy.float(1 + numpy.median(rsflsf))
                minw = (rlmed - rlmin) / rlmed
                ktab[rc,lc] = (minw * rlmin + (1 - minw) * rlmed) * (nrl * nrl / (nrf * nlf))

        # add "kerning additions"
        ktab = ktab - self._xktab

        # overall median
        ktmed = numpy.median(ktab[numpy.isfinite(ktab)]) + 0.1 * (
            numpy.float(self._image.shape[1]) / numpy.float(nl))
        ktab = ktmed - ktab
        ktab[numpy.isnan(ktab)] = 0.0

        # adjust "space" kerning
        ktabsp = numpy.ceil(numpy.mean(ktab[1:,1:]))
        ktab[0,:] = -ktabsp
        ktab[:,0] = -ktabsp

        # store table
        self._kerning = numpy.trunc(ktab).astype(numpy.int32)

    # set single line into images
    def set_line(self,
        line:Union[str,list],
        fsize:float,
        spkern:int = 0,
        xkern:int = 0,
        ) -> numpy.ndarray:

        # IMPORTS DONE HERE TO SAVE TIME AT MODULE INIT
        from .sampler import Sampler
        sampler = Sampler()

        if isinstance(line, str):
            line = [line]
        elif not isinstance(line, list):
            raise ValueError('Invalid line(s) to set.')
        out = [None] * len(line)
        if fsize < 1.0:
            raise ValueError('Invalid fsize parameter.')
        ffact = fsize / numpy.float(self._size)
        ifsize = self._image.shape
        for lc in range(len(line)):
            letters = [ord(l) for l in line[lc]]
            nletters = len(letters)
            if nletters == 0:
                out[lc] = numpy.zeros(0, dtype=numpy.uint8).reshape((fsize,0,))
                continue
            let_ims = [None] * nletters
            let_spc = numpy.zeros(nletters, dtype=numpy.int32)
            for letc in range(nletters):
                leti = self._lmap[letters[letc]]
                if leti < 0:
                    leti = self._lmap[63]
                let_ims[letc] = self._image[:, self._xstart[leti]:self._xstop[leti]]
                if letc < (nletters - 1):
                    nleti = self._lmap[letters[letc+1]]
                    if nleti < 0:
                        nleti = self._lmap[63]
                    if nleti == 0:
                        let_spc[letc] = self._kerning[leti,nleti] + spkern
                    else:
                        let_spc[letc] = self._kerning[leti,nleti] + xkern
            
            # element and total size
            xsims = numpy.asarray([l.shape[1] for l in let_ims])
            xstot = numpy.sum(xsims) + numpy.sum(let_spc)
            lineimage = numpy.zeros(ifsize[0] * xstot, dtype=numpy.uint8).reshape(
                (ifsize[0], xstot,))
            lii = 0
            for letc in range(nletters):
                lineimage[:,lii:lii+xsims[letc]] = numpy.maximum(
                    lineimage[:,lii:lii+xsims[letc]], let_ims[letc])
                lii += xsims[letc] + let_spc[letc]
            
            # resize
            if ffact != 1.0:
                lineimage = sampler.sample_grid(lineimage, ffact, 'resample', 'uint8')

            # store
            out[lc] = lineimage
        return out
    
    # set text into image
    def set_text(self,
        text:str,
        fsize:float = 24.0,
        color:list = [0, 0, 0],
        bcolor:list = [255, 255, 255],
        align:str = 'left',
        invert:bool = False,
        outsize_x:int = 0,
        outsize_y:int = 0,
        padding:int = 4,
        spkern:int = 0,
        xkern:int = 0,
        ) -> Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:

        if not isinstance(text, str):
            raise ValueError('Invalid text.')
        text = text.split('\n')
        if not isinstance(fsize, float) or fsize <= 1.0:
            fsize = 24.0
        fsize = numpy.ceil(fsize)
        if not isinstance(bcolor, list) or len(bcolor) != 3:
            bcolor = [255, 255, 255]
        try:
            bcolor = numpy.asarray(bcolor).astype(numpy.uint8)
        except:
            raise
        if not isinstance(color, list) or len(color) != 3:
            color = [0, 0, 0]
        try:
            color = numpy.asarray(color).astype(numpy.uint8)
        except:
            raise
        if not isinstance(invert, bool):
            invert = False
        if not isinstance(outsize_x, int) or outsize_x < 0:
            outsize_x = 0
        if not isinstance(outsize_y, int) or outsize_y < 0:
            outsize_y = 0
        if not isinstance(padding, int) or padding < 0:
            padding = 0
        if not isinstance(spkern, int) or spkern < -8 or spkern > 48:
            spkern = 0
        if not isinstance(xkern, int) or xkern < -16 or xkern > 16:
            xkern = 0

        # set each line with current settings
        lines = self.set_line(text, fsize, spkern, xkern)
        padsz = numpy.round_([2 * padding])[0]
        fsize0 = numpy.asarray([line.shape[0] for line in lines])
        fstot0 = numpy.sum(fsize0) + padsz
        fsize1 = numpy.asarray([line.shape[1] for line in lines])
        fstot1 = numpy.amax(fsize1) + padsz

        # outsize needs to be determined
        if outsize_y == 0:
            outsize_y = fstot0
            ypad = padding
        else:
            ypad = 0
        if outsize_x == 0:
            outsize_x = fstot1
            xpad = padding
        else:
            xpad = 0

        # create image
        if outsize_y >= fstot0:
            if outsize_x >= fstot1:
                ima = numpy.zeros(outsize_y * outsize_x,
                    dtype=numpy.float32).reshape((outsize_y, outsize_x,))
            else:
                ima = numpy.zeros(outsize_y * fstot1,
                    dtype=numpy.float32).reshape((outsize_y, fstot1,))
        else:
            if outsize_x >= fstot1:
                ima = numpy.zeros(fstot0 * outsize_x,
                    dtype=numpy.float32).reshape((fstot0, outsize_x,))
            else:
                ima = numpy.zeros(fstot0 * fstot1,
                    dtype=numpy.float32).reshape((fstot0, fstot1,))
        imash = ima.shape
        im = numpy.zeros(imash[0] * imash[1] * 3,
            dtype=numpy.uint8).reshape((imash[0], imash[1], 3,))
        for pc in range(3):
            im[:,:,pc] = bcolor[pc]

        # store font pieces into
        yfrom = ypad
        yfroms = numpy.zeros(len(lines), dtype=numpy.int32)
        for lc in range(len(lines)):
            lim = (1.0 / 255.0) * lines[lc].astype(numpy.float32)
            if lim.shape[1] > outsize_x:
                lim = lim[:, 0:outsize_x]
            if invert:
                lim = 1.0 - lim
            tsize = lim.shape[1]
            if align == 'left':
                xfrom = xpad
            else:
                xleft = imash[1] - 2 * xpad
                if align == 'right':
                    xfrom = xpad + xleft - tsize
                else:
                    xfrom = xpad + ((xleft - tsize) // 2)
            ima[yfrom:yfrom+fsize0[lc], xfrom:xfrom+tsize] = lim
            lim = (lim > 0.0).astype(numpy.float32)
            for pc in range(3):
                cim = im[yfrom:yfrom+fsize0[lc], xfrom:xfrom+tsize, pc]
                im[yfrom:yfrom+fsize0[lc], xfrom:xfrom+tsize, pc] = numpy.trunc(
                    0.5 + (1.0 - lim) * cim.astype(numpy.float32) +
                    numpy.float(color[pc]) * lim).astype(numpy.uint8)
            yfroms[lc] = yfrom
            yfrom = yfrom + fsize0[lc]
        
        # resize total if needed
        if imash[0] > outsize_y or imash[1] > outsize_x:
            o_shape = (outsize_y, outsize_x)
            from .sampler import Sampler
            sampler = Sampler()
            ima = sampler.sample_grid(ima, o_shape, 'resample', 'float32')
            im = sampler.sample_grid(im, o_shape, 'resample', 'uint8')
        
        # return
        return (im, ima, yfroms)



"""
isicarchive.image (Image)

This module provides the Image object for the IsicApi to utilize.

Image objects are either returned from calls to

   >>> from isicarchive.api import IsicApi
   >>> api = IsicApi()
   >>> image = api.image(image_id)

or can be generated

   >>> from isicarchive.image import Image
   >>> image = Image(...)
"""

__version__ = '0.4.8'


# imports (needed for majority of functions)
import datetime
import glob
import os
from typing import Any, List, Union
import warnings

from . import func
from .vars import ISIC_IMAGE_DISPLAY_SIZE_MAX

_json_full_fields = [
    'created',
    'creator',
    'dataset',
    'id',
    'meta',
    'name',
    'notes',
    'updated',
]
_json_partial_fields = [
    'id',
    'name',
    'updated',
]
_mangling = {
    'id': '_id',
}
_repr_pretty_list = {
    'id': 'id',
    'name': 'name',
    'dataset_id': 'dataset._id',
    'dataset_name': 'dataset.name',
    'dataset_description': '_dataset.description',
    'dataset_license': '_dataset.license',
    'meta_acquisition': 'meta.acquisition.{}',
    'meta_clinical': 'meta.clinical.{keys}',
    'meta_clinical_benign_malignant': 'meta.clinical.benign_malignant',
    'meta_clinical_diagnosis': 'meta.clinical.diagnosis',
    'meta_clinical_diagnosis_confirm_type': 'meta.clinical.diagnosis_confirm_type',
    'meta_clinical_melanocytic': 'meta.clinical.melanocytic',
    'notes_reviewed': 'notes.reviewed.{}',
    'superpixels_max': 'superpixels.max',
    'superpixels_shape': 'superpixels.shp',
}

class Image(object):
    """
    Image object. If the details are not filled in, only the `id`, `name`,
    and `updated` fields will be set.

    To generate an image object for an existing dataset, please use the
    IsicApi.image(...) method!

    To generate a new image object (for later storage), use

       >>> image = Image(name=image_name, data=...)

    Attributes
    ----------
    created : Date
        Image creation date (w.r.t. in the database!)
    creator : dict
        Contains _id and (short) name field identifying the creator
    data : numpy.ndarray (or None)
        Once data is loaded, the imread decoded image will be in this field
    dataset : dict
        Dataset fields (non-detailed)
    id : str
        mongodb objectId of the image
    meta : dict
        Metadata associated with the image, containing fields
        'acquisition': dict (image type, size)
        'clinical': dict (patient/lesion classification)
        'private': dict (original filename, etc.)
        'unstructured': dict (additional metadata fields)
        'unstructuredExif': dict (additional image metadata fields)
    name : str
        ISIC archive filename (generally ISIC_#######, without extension!)
    notes : dict
        Notes (settings) within the archive, contains:
        'reviewed': dict (when was the image reviewed)
    updated : Date
        Image update date (w.r.t. in the database!)
    
    Methods
    -------
    """


    def __init__(self,
        from_json:dict = None,
        name:str = None,
        api:object = None,
        load_image_data:bool = False,
        load_superpixels:bool = False,
        ):
        """Image init."""

        self._api = api
        self._dataset = None
        self._detail = False
        self._in_archive = False
        self._model_type = 'image'
        self._raw_data = None
        self._segmentation = None
        # still needs timezone information!!
        self.created = datetime.datetime.now().strftime(
            '%Y-%m-%dT%H:%M:%S.%f+00:00')
        self.creator = {'_id': '000000000000000000000000'}
        self.data = None
        self.dataset = dict()
        self.id = ''
        self.meta = dict()
        self.name = name if name else 'ISIC_xxxxxxxx'
        self.notes = dict()
        self.superpixels = {
            'idx': None,
            'map': None,
            'max': 0,
            'shp': (0, 0),
            'spd': None,
            'szp': None,
            'szs': None,
            'xyc': None,
        }
        self.updated = self.created

        # from JSON
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json)
            except:
                raise
        if self._in_archive and load_image_data:
            try:
                self.load_image_data()
            except Exception as e:
                warnings.warn(str(e))
        if self._in_archive and load_superpixels:
            try:
                self.load_superpixels(map_superpixels=True)
            except Exception as e:
                warnings.warn(str(e))

    # parse JSON
    def _from_json(self, from_json:dict):
        self.id = from_json['_id']
        self.name = from_json['name']
        self.updated = from_json['updated']
        if 'creator' in from_json:
            self.created = from_json['created']
            self.creator = from_json['creator']
            if 'dataset' in from_json:
                self.dataset = from_json['dataset']
                if '_id' in self.dataset and (self.dataset['_id'] in
                    self._api._datasets):
                    self._dataset = self._api._datasets[self.dataset['_id']]
            if 'meta' in from_json:
                self.meta = from_json['meta']
            if 'notes' in from_json:
                self.notes = from_json['notes']
            self._detail = True
        self._in_archive = True

    # JSON
    def __repr__(self):
        return 'isicarchive.image.Image(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Image "{0:s}" (id={1:s}, image data loaded: {2:s})'.format(
            self.name, self.id, str(not self.data is None))
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)
    
    # JSON representation (without constructor):
    def as_json(self):

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        from json import dumps as json_dumps

        json_list = []
        fields = _json_full_fields if self._detail else _json_partial_fields
        for field in fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json_dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'

    # clear data
    def clear_data(self,
        clear_raw_data:bool = True,
        clear_data:bool = True,
        clear_superpixels:bool = True,
        clear_segmentation:bool = True,
        deref_dataset:bool = False,
        deref_segmentation:bool = False,
        ):
        if deref_dataset:
            self._dataset = None
        if clear_raw_data:
            self._raw_data = None
        if clear_data:
            self.data = None
        if clear_superpixels:
            self.superpixels = {
                'idx': None,
                'map': None,
                'max': 0,
                'shp': (0, 0),
                'spd': None,
                'szp': None,
                'szs': None,
                'xyc': None,
            }
        if clear_segmentation:
            if self._segmentation:
                self._segmentation.clear_data(deref_image=deref_segmentation)
        if deref_segmentation:
            self._segmentation = None

    # load image data
    def load_image_data(self, keep_raw_data:bool = False):

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        import imageio

        if not self._api:
            raise ValueError('Invalid image object to load image data for.')
        if self._api._cache_folder:
            image_filename = self._api.cache_filename(self.id, 'image', '.*', '*')
            image_list = glob.glob(image_filename)
            if image_list:
                try:
                    self.data = None
                    with open(image_list[0], 'rb') as image_file:
                        image_raw = image_file.read()
                    if keep_raw_data:
                        self._raw_data = image_raw
                    if (image_list[0][-4:].lower() == '.jpg' or
                        image_list[0][-5:].lower() == '.jpeg'):
                        self.data = imageio.imread(image_raw, exifrotate=False)
                    else:
                        self.data = imageio.imread(image_raw)
                    return
                except Exception as e:
                    warnings.warn('Error loading image: ' + str(e))
                    os.remove(image_list[0])
        if self._in_archive and self._api:
            try:
                req = self._api.get('image/' + self.id + '/download',
                    parse_json=False)
                if req.ok:
                    image_raw = req.content
                    image_type = func.guess_file_extension(req.headers)
                    if keep_raw_data:
                        self._raw_data = image_raw
                    if image_type == '.jpg':
                        self.data = imageio.imread(image_raw, exifrotate=False)
                    else:
                        self.data = imageio.imread(image_raw)
                    if self._api._cache_folder:
                        if self.name and (len(self.name) > 5):
                            extra = self.name
                        else:
                            extra = None
                        image_filename = self._api.cache_filename(self.id,
                            'image', func.guess_file_extension(req.headers),
                            extra)
                        with open(image_filename, 'wb') as image_file:
                            image_file.write(image_raw)
            except Exception as e:
                warnings.warn('Error loading image data: ' + str(e))

    # load segmentation
    def load_segmentation(self):
        if not self._api:
            return
        try:
            if self._segmentation:
                seg_obj = self._segmentation
            else:
                seg_obj = self._api.segmentation(self.id)
                self._segmentation = seg_obj
        except:
            return
        seg_obj.load_mask_data()

    # load image superpixels
    def load_superpixels(self, map_superpixels:bool = False):

        # IMPORTS DONE HERE TO SAVE TIME AT MODULE INIT
        import imageio
        import numpy
        from .jitfunc import superpixel_decode

        if not self.superpixels['idx'] is None:
            return
        if not self._api:
            raise ValueError('Invalid image object to load superpixels for.')
        spimg_filename = self._api.cache_filename(self.id, 'spimg', '.png')
        if self._api._cache_folder and os.path.exists(spimg_filename):
            try:
                with open(spimg_filename, 'rb') as image_file:
                    image_raw = image_file.read()
                image_png = imageio.imread(image_raw)
                self.superpixels['idx'] = superpixel_decode(image_png)
                self.superpixels['max'] = numpy.amax(
                    self.superpixels['idx']).item()
                self.superpixels['shp'] = self.superpixels['idx'].shape
            except Exception as e:
                os.remove(spimg_filename)
                warnings.warn('Error loading image: ' + str(e))
        if self.superpixels['idx'] is None:
            try:
                req = self._api.get('image/' + self.id + '/superpixels',
                    parse_json=False)
                if req.ok:
                    image_raw = req.content
                    image_png = imageio.imread(image_raw)
                    if self._api._cache_folder:
                        with open(spimg_filename, 'wb') as image_file:
                            image_file.write(image_raw)
                    self.superpixels['idx'] = superpixel_decode(image_png)
                    self.superpixels['max'] = numpy.amax(
                        self.superpixels['idx']).item()
                    self.superpixels['shp'] = self.superpixels['idx'].shape
                else:
                    raise RuntimeError('HTTP server error: ' + req.text)
            except Exception as e:
                warnings.warn('Error loading superpixels: ' + str(e))
                return
        if not map_superpixels:
            return
        try:
            if self.superpixels['map'] is None:
                self.map_superpixels()
        except Exception as e:
            warnings.warn('Error processing superpixel data: ' + str(e))

    # map superpixels
    def map_superpixels(self):

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        import numpy
        from .jitfunc import superpixel_map

        if not self.superpixels['map'] is None:
            return
        clear_seg = False
        try:
            if self.superpixels['idx'] is None:
                self.load_superpixels()
            if self.superpixels['idx'] is None:
                raise ValueError('Some problem occurred during load_superpixels().')
            if self._segmentation is None:
                try:
                    clear_seg = True
                    self.load_segmentation()
                except:
                    pass
        except Exception as e:
            warnings.warn('Error loading superpixels: ' + str(e))
            return
        pixel_img = self.superpixels['idx']
        try:
            pi_cols = pixel_img.shape[1]
            sp_map = superpixel_map(pixel_img)
            sp_shp = sp_map.shape
            self.superpixels['map'] = sp_map
            self.superpixels['szs'] = sp_map[:,-1].tolist()
            if not self._segmentation is None and not self._segmentation.mask is None:
                sim = self._segmentation.superpixels_in_mask()
                a = self._segmentation.area
                self.superpixels['szp'] = [
                    p * s / a for (p,s) in zip(sim, self.superpixels['szs'])]
            self.superpixels['xyc'] = [None] * sp_shp[0]
            for spidx in range(sp_shp[0]):
                splen = sp_map[spidx,-1]
                spcrd = sp_map[spidx,0:splen]
                spcrx = spcrd % pi_cols
                spcry = spcrd // pi_cols
                self.superpixels['xyc'][spidx] = [
                    int(numpy.trunc(numpy.mean(spcrx)) + 0.5),
                    int(numpy.trunc(numpy.mean(spcry)) + 0.5)]
        except Exception as e:
            warnings.warn('Error mapping superpixels: ' + str(e))
        if clear_seg:
            try:
                self._segmentation.clear_data()
            except:
                pass

    # mark superpixels
    def mark_superpixels(self, edge_width:int = 1, color:list = [0,0,0]):
        if self.data is None:
            self.load_image_data()
        if self.superpixels['map'] is None:
            self.map_superpixels()
        outlines = self.superpixel_outlines('coords')
        image_data = self.data.copy()
        im_shape = image_data.shape
        if len(im_shape) > 2:
            p = im_shape[2]
        else:
            p = 1
        image_data = image_data.reshape((im_shape[0] * im_shape[1], p,))
        image_x = self.data.shape[1]
        for sppcrd in outlines.values():
            sppcrd = image_x * sppcrd[:,0] + sppcrd[:,1]
            image_data[sppcrd, 0] = color[0]
            image_data[sppcrd, 1] = color[1]
            image_data[sppcrd, 2] = color[2]
        self.data = image_data.reshape(im_shape)

    # create a meta-information dict with all kinds of information
    def meta_info(self,
        data_info:bool = True,
        metadata_info:bool = True,
        segmentation_info:bool = True,
        superpixel_info:bool = True,
        ) -> dict:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        import numpy
        from .imfunc import segmentation_outline

        if not self._api:
            warnings.warn('Full results only available with API connection.')
        info = dict()
        info['_id'] = self.id
        info['name'] = self.name
        info['dataset'] = self.dataset
        if data_info:
            self.load_image_data()
            info['image'] = dict()
            image_data = self.data
            imshape = image_data.shape
            r = int(numpy.mean(image_data[:,:,0]))
            g = int(numpy.mean(image_data[:,:,1]))
            b = int(numpy.mean(image_data[:,:,2]))
            info['image']['height'] = int(imshape[0])
            info['image']['width'] = int(imshape[1])
            info['image']['num_pixels'] = int(imshape[0] * imshape[1])
            info['image']['RGB_average'] = [r, g, b]
        if metadata_info:
            info['meta'] = self.meta
        if segmentation_info:
            self.load_segmentation()
            info['segmentation'] = dict()
            seg_mask = (self._segmentation.mask > 0)
            info['segmentation']['pixels_in_mask'] = int(numpy.sum(seg_mask))
            info['segmentation']['osvgp_mask'] = segmentation_outline(seg_mask,
                negative=False)
            info['segmentation']['osvgp_hole'] = segmentation_outline(seg_mask,
                negative=True)
            sp_in_mask = self._segmentation.superpixels_in_mask()
        else:
            sp_in_mask = None
        if superpixel_info:
            self.load_superpixels()
            self.map_superpixels()
            info['superpixels'] = dict()
            info['superpixels']['number'] = int(self.superpixels['max']) + 1
            info['superpixels']['cjson'] = self.superpixel_outlines('cjson')
            info['superpixels']['osvgp'] = self.superpixel_outlines('osvgp')
            info['superpixels']['numpix'] = self.superpixels['szs']
            info['superpixels']['centers'] = self.superpixels['xyc']
            if not sp_in_mask is None:
                info['superpixels']['sp_in_mask'] = sp_in_mask
        self.clear_data()
        return info

    # POST metadata for an image to the image/{id}/metadata endpoint
    def post_metadata(self,
        metadata:dict,
        ) -> bool:
        """
        POSTs metadata for an image and returns True or False
        """
        if (not self._api) or (not self._api._auth_token):
            raise ValueError('Invalid image object to post metadata with.')
        if not func.could_be_mongo_object_id(self.id):
            raise ValueError('Invalid image object_id format.')
        req = self._api.post('image/' + self.id + '/metadata',
            params={'metadata': metadata, 'save': 'true'}, parse_json=False)
        if not req.ok:
            warnings.warn("Image metadata posting failed: " + req.text)
            return ''
        return req.json()

    # retrieve associated segmentations
    def segmentations(self,
        as_objects:bool = True,
        ) -> list:
        pass

    # show image in notebook
    def show_in_notebook(self,
        max_size:int = None,
        library:str = 'matplotlib',
        call_display:bool = True,
        ) -> object:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        from .imfunc import display_image
        try:
            if self.data is None:
                self.load_image_data()
            return display_image(self.data, max_size=max_size,
                ipython_as_object=(not call_display), library=library)
        except Exception as e:
            warnings.warn('show_in_notebook(...) failed: ' + str(e))

    # superpixel neighbors
    def superpixel_neighbors(self,
        up_to_degree:int = 1) -> List:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        from .imfunc import superpixel_neighbors
        
        if self.superpixels['map'] is None:
            try:
                self.map_superpixels()
                if self.superpixels['map'] is None:
                    raise RuntimeError('Error mapping superpixels.')
            except:
                raise
        try:
            return superpixel_neighbors(self.superpixels['idx'],
                self.superpixels['map'], up_to_degree)
        except:
            raise

    # superpixel outlines
    def superpixel_outlines(self,
        out_format:str = 'osvgp',
        pix_selection:list = None,
        path_attribs:Union[list,str] = None,
        ) -> Any:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE IMPORT
        from .imfunc import superpixel_outlines

        if self.superpixels['map'] is None:
            self.map_superpixels()
            if self.superpixels['map'] is None:
                warnings.warn('Could not load or process superpixel data.')
                return None
        if self.superpixels['idx'] is None:
            self.load_superpixels()
            if self.superpixels['idx'] is None:
                warnings.warn('Could not load or process superpixel data.')
                return None
        outlines = superpixel_outlines(
            self.superpixels['map'], self.superpixels['idx'].shape,
            out_format=out_format, pix_selection=pix_selection,
            path_attribs=path_attribs)
        if out_format == 'osvgp' and pix_selection is None:
            self.superpixels['spd'] = outlines
        return outlines

    # shortcut to save a superpixel JSON file
    def superpixel_savejson(self, filename:str):

        # IMPORT DONE HERE TO SAVE TIME AT MODULE IMPORT
        from json import dumps as json_dumps
        from .imfunc import superpixel_outlines

        contours = self.superpixel_outlines('cjson')
        json_str = json_dumps(contours) + "\n"
        try:
            with open(filename, 'w') as json_file:
                json_file.write(json_str)
        except:
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 30 11:01:59 2019

@author: weberj3
"""

import os
from matplotlib import pyplot
#%matplotlib inline
#import imageio
import numpy
from isicarchive import imfunc
from isicarchive.api import IsicApi
from isicarchive import sampler

# function for mean and sample STD
#def mean_std(a:list, is_sample:bool=True):
#    ddof = 1 if is_sample else 0
#    return (numpy.mean(a), numpy.std(a, ddof=ddof))

# set to True if you would like (more) details in the print-out
#print_details = False
#print_fine_details = False

# please change the username accordingly!
username = 'weberj3@mskcc.org'

# root folder for all ISIC related data
doc_folder = 'Z:\\10.Imaging Informatics\\'

# cache folder
cache_folder = doc_folder + 'ISIC' + os.sep + 'cache'

# show URL requests (for debugging purposes only!)
debug = False

# instantiate API object
api = IsicApi(username, cache_folder=cache_folder, debug=debug)

# load image
image = api.image('ISIC_0023044')
image.load_image_data()
gray = imfunc.image_gray(image.data)
gray_0 = gray[:,:,0]
#im_t = imfunc.read_image('C:\\Users\weberj3\\Documents\\testx.png').astype(numpy.float64)
#im_s = imfunc.read_image('C:\\Users\weberj3\\Documents\\testxr.png').astype(numpy.float64)
#im_t = (im_t - 32.0)
#im_s = (im_s - 32.0)

s = sampler.Sampler()
m = {
    'origin': 0.5 * numpy.asarray(list(gray_0.shape), numpy.float64),
    'rotate': numpy.asarray([0.1], numpy.float64),
    'trans': numpy.asarray([20.0,30.0], numpy.float64),
}
#t = sampler.trans_matrix(m)
#print(t)

r_image = s.sample_grid(gray_0, 1.0, 'cubic', out_type='uint8', m=m)

#import scipy.ndimage as ndi
#mask = ndi.binary_dilation(r_image >= 1)

fig = pyplot.figure(figsize=(8,12))
pyplot.imshow(r_image)
pyplot.show()

mx = imfunc.image_register(gray_0, r_image)
print(mx)

m2 = {
    'origin': 0.5 * numpy.asarray(list(r_image.shape), numpy.float64),
    'rotate': mx[1],
    'trans': mx[0],
}
r_image_back = s.sample_grid(r_image, 1.0, 'cubic', out_type='uint8', m=m2)

fig = pyplot.figure(figsize=(8,12))
yo = gray_0.shape[0] // 10
xo = gray_0.shape[1] // 10
pyplot.imshow(r_image_back[yo:-yo, xo:-xo].astype(numpy.float64) -
              gray_0[yo:-yo, xo:-xo].astype(numpy.float64))
pyplot.show()

#image.load_image_data()
#image.load_superpixels()
#image.map_superpixels()
#nei = imfunc.superpixel_neighbors(image.superpixels['idx'], image.superpixels['map'], 3)
#print(nei)

# study folder
#study_folder = doc_folder + 'EASY' + os.sep + 'PILOT' + os.sep

# load study and data
#study = api.study('ISIC Annotation Study - All Features')
#study.cache_image_data()
#study.load_annotations()
# load meta data
#meta_data_url = ('https://raw.githubusercontent.com/neuroelf/' +
 #   'isicarchive/master/data/EASY_pilot_diagnoses.csv')
#study.load_meta_data(meta_data_url, list_to_dict=True,
#    dict_key='name', extract_key=['diagnosis', 'exemplar'])

## and create a dictionary mapping diagnosis to a list of images
#diag_images = dict()
#for (name, diag) in study.meta_data['diagnosis'].items():
#    if not diag in diag_images:
#        diag_images[diag] = []
#    diag_images[diag].append(name)
#
## same for exemplar features
#exem_images = dict()
#for (name, exemplar) in study.meta_data['exemplar'].items():
#    if not exemplar:
#        continue
#    if not exemplar in exem_images:
#        exem_images[exemplar] = []
#    exem_images[exemplar].append(name)

# select only from users that completed the study
#num_images = len(study.images)
#users = [u for (u,c) in study.user_completion.items() if c==num_images]

## create heatmaps with default settings (if not yet done)
#study_stats_file = study_folder + 'heatmap_stats.json.gz'
#if not os.path.exists(study_stats_file):
#    mix_colors = False
#    underlay_gray = 0.8
#    study_stats = study.image_heatmaps(study_folder, users=users,
#        mix_colors=mix_colors, underlay_gray=underlay_gray)
#else:
#    study_stats = func.gzip_load_var(study_stats_file)
#
#
## select those annotations, and gather basic statistics
#study.select_annotations(users=users)
#total_features_annotations = sum(
#    [len(a.features) for a in study.annotation_selection.values()])
#selected_features = dict()
#for annotation in study.annotation_selection.values():
#    for feature in annotation.features:
#        selected_features[feature] = True

#overlap_stats = study.overlap_stats(users=users)



"""
isicarchive.statsimage (StatsImage)

This module provides the StatsImage class and associated methods.
"""

__version__ = '0.4.8'


from typing import Any, Tuple, Union
import warnings

from numba import jit
import numpy

from . import colorlut
from .imfunc import image_mix, lut_lookup


class StatsLayer(object):
    """
    StatsLayer
    """


    def __init__(self,
        data:Union[tuple, numpy.ndarray],
        alpha:Union[numpy.ndarray, float, None] = None,
        lut_name:Union[str, list] = None,
        min_thresh:float = 0.1,
        max_thresh:float = 1.0,
        min_alpha:float = 0.25,
        max_alpha:float = 1.0,
        show_neg:bool = False,
        ):
        if data is None:
            raise ValueError('Requires data argument.')
        if isinstance(data, tuple):
            try:
                data = numpy.zeros(data[0] * data[1],
                    dtype=numpy.float32).reshape(data)
            except:
                raise ValueError('Invalid data shape.')
        shape = data.shape
        if len(shape) != 2:
            raise ValueError('Invalid data shape')
        if not alpha is None:
            if isinstance(alpha, float):
                alpha = numpy.float32(alpha) * numpy.ones(shape[0] * shape[1],
                    dtype=numpy.float32).reshape(shape)
            elif alpha.shape != shape:
                raise ValueError('Alpha layer shape must match.')
        else:
            alpha = numpy.zeros(shape[0] * shape[1],
                dtype=numpy.float32).reshape(shape)
        self._rendered = None
        self._rendered_a = None
        self._shape = shape
        self.alpha = alpha
        self.data = data
        if isinstance(lut_name, str) and (lut_name in colorlut.LUTs):
            self.lut = colorlut.LUTs[lut_name]
        else:
            self.lut = colorlut.LUTs['standard']
        for (idx, l) in enumerate(self.lut):
            self.lut[idx] = numpy.asarray(l, dtype=numpy.uint8)
        if len(self.lut) < 2:
            self.lut.append(None)
        self.max_alpha = max_alpha
        self.max_thresh = max_thresh
        self.min_alpha = min_alpha
        self.min_thresh = min_thresh
    
    def _render(self):
        if self.min_thresh <= 0.0:
            self.min_thresh = 0.000000001
        if self.max_thresh <= self.min_thresh:
            self.max_thresh = self.min_thresh + 0.000001
        abs_data = (numpy.abs(self.data) - self.min_thresh) / (
            self.max_thresh - self.min_thresh)
        show_data = (abs_data >= self.min_thresh)
        self._rendered = lut_lookup(self.data, self.lut[0], self.lut[1],
            self.max_thresh - self.min_thresh, self.min_thresh)
        self._rendered_a = numpy.zeros(self._shape[0] * self._shape[1],
            dtype=numpy.float32).reshape(self._shape)
        self._rendered_a[show_data] = numpy.minimum(1.0, self.min_alpha + 
            (self.max_alpha - self.min_alpha) * abs_data[show_data])
    def rendered(self):
        if self._rendered is None:
            self._render()
        return (self._rendered, self._rendered_a)


class StatsImage(object):
    """
    StatsImage

    Attributes
    ----------
    shape : tuple
        2-element array with Y (height) and X (width) elements
    stats : list of dicts
        Statistics layers with mandatory fields
        stats['data'] : ndarray (the statistical data)
        stats['lut'] : Cx3 color lookup table or str with name
    underlay : ndarray
        Either Grayscale or RGB underlay image, can be none
    """


    def __init__(self,
        underlay:numpy.ndarray = None,
        stats_data:numpy.ndarray = None,
        stats_lut:Any = None,
        shape:Tuple = None,
        ):
        self._rendered = None
        if not underlay is None:
            shape = underlay.shape
        if len(shape) > 2:
            shape = (shape[0], shape[1])
        self.shape = shape
        self.underlay = underlay
        self.stats = []
        if not stats_data is None:
            self.stats.append(StatsLayer(stats_data, stats_lut))
    
    def render(self):
        if self.underlay is None:
            pass
        pass
    
    def rendered(self):
        if not self._rendered:
            self.render()
        return self._rendered



"""
isicarchive.sampler (Sampler)

This module provides the Sampler helper class and doesn't have to be
imported from outside the main package functionality (IsicApi).
"""

# specific version for file
__version__ = '0.4.8'


# imports (needed for majority of functions)
from typing import Any, List, Union
import warnings

from numba import float64, int64, jit, prange
import numpy


# fix value to -1.0 ... 1.0
def _frone(v):
    return min(1.0, max(-1.0, v))

# Gaussian (smoothing) kernel
def _gauss_kernel(fwhm:numpy.float64 = 2.0) -> numpy.ndarray:
    if fwhm <= 0.29:
        return numpy.asarray([0,1,0]).astype(numpy.float64)
    fwhm = fwhm / numpy.sqrt(8.0 * numpy.log(2.0))
    if fwhm < 2.0:
        md = numpy.trunc(0.5 + 6.0 * fwhm)
    else:
        md = numpy.trunc(0.5 + 6.0 * numpy.log2(fwhm) * fwhm)
    k = numpy.exp(-((numpy.arange(-md,md+1.0,1.0) ** 2) / (2.0 * fwhm * fwhm)))
    k = k[k >= 0.00000000001]
    return k / numpy.sum(k)

# sample grid
@jit([
    'f8[:,:](f8[:,:],f8[:],f8[:],f8[:],i8)', #output(array, crd0, crd1, kernel, ksize)
    'f8[:,:](f4[:,:],f8[:],f8[:],f8[:],i8)',
    'f8[:,:](u1[:,:],f8[:],f8[:],f8[:],i8)',
    ], nopython=True)
def _sample_grid_2d(
    a:numpy.ndarray,
    c0:numpy.ndarray,
    c1:numpy.ndarray,
    k:numpy.ndarray,
    ks:int64) -> numpy.ndarray:
    nk = k.size - 1
    kl = float(nk) / float(2 * ks)
    if kl != numpy.trunc(kl):
        raise ValueError('Invalid kernel.')
    ikl = int(kl)
    mks = ikl * ks
    fks = float(ks)
    as0 = a.shape[0]
    as1 = a.shape[1]
    nc0 = c0.size
    nc1 = c1.size
    mn1 = int(numpy.amin(c1) - (kl + 1.0))
    if mn1 < 0:
        mn1 = 0
    mx1 = int(numpy.amax(c1) + (kl + 1.0))
    if mx1 > as1:
        mx1 = as1
    l1 = mx1 - mn1
    out = numpy.zeros(nc0 * nc1, dtype=numpy.float64).reshape((nc0,nc1,))
    for i0 in prange(nc0): #pylint: disable=not-an-iterable
        c0c = c0[i0]
        c0b = int(c0c + 0.5)
        c0o = c0c - float(c0b)
        row = numpy.zeros(l1, dtype=numpy.float64).reshape((1,l1,))
        rw = 0.0
        for ri in range(c0b-ikl, c0b+ikl+1):
            if ri < 0 or ri >= as0:
                continue
            wi = mks + (ri-c0b) * ks - int(c0o * fks)
            if wi > 0 and wi < nk:
                kwi = k[wi]
                row += kwi * a[ri, mn1:mx1].astype(numpy.float64)
                rw += kwi
        if rw == 0.0:
            rw = 1.0
        row /= rw
        ascol = row.reshape(l1,)
        for i1 in range(nc1):
            c1c = c1[i1]
            c1b = int(c1c + 0.5 - mn1)
            c1o = c1c - float(c1b)
            val = 0.0
            vw = 0.0
            for ci in range(c1b-ikl, c1b+ikl+1):
                if ci < 0 or ci >= l1:
                    continue
                cwi = mks + (ci-c1b) * ks - int(c1o * fks)
                if cwi > 0 and cwi < nk:
                    kwi = k[cwi]
                    valp = ascol[ci]
                    if not numpy.isnan(valp):
                        val += kwi * valp
                        vw += kwi
            if vw == 0.0:
                vw = 1.0
            out[i0,i1] = val / vw
    return out

# sample grid coordinates
@jit([
    'f8[:](f8[:,:],f8[:,:],f8[:],i8)', #output(array, crd, kernel, ksize)
    'f8[:](f4[:,:],f8[:,:],f8[:],i8)',
    'f8[:](u1[:,:],f8[:,:],f8[:],i8)',
    ], nopython=True)
def _sample_grid_coords(
    a:numpy.ndarray,
    c:numpy.ndarray,
    k:numpy.ndarray,
    ks:int64) -> numpy.ndarray:
    nk = k.size - 1
    kl = float(nk) / float(2 * ks)
    if kl != numpy.trunc(kl):
        raise ValueError('Invalid kernel.')
    ikl = int(kl)
    mks = ikl * ks
    fks = float(ks)
    as0 = a.shape[0]
    as1 = a.shape[1]
    nc = c.shape[0]
    if c.shape[1] != 2:
        raise ValueError('Invalid coordinate list.')
    out = numpy.zeros(nc, dtype=numpy.float64).reshape((nc,))
    for i in prange(nc): #pylint: disable=not-an-iterable
        c0c = c[i,0]
        c0b = int(c0c + 0.5)
        c0o = c0c - float(c0b)
        c1c = c[i,1]
        c1b = int(c1c + 0.5)
        c1o = c1c - float(c1b)
        val = 0.0
        vw = 0.0
        for ri in range(c0b-ikl, c0b+ikl+1):
            if ri < 0 or ri >= as0:
                continue
            wi = mks + (ri-c0b) * ks - int(c0o * fks)
            if wi < 0 or wi >= nk:
                continue
            kwi0 = k[wi]
            for ci in range(c1b-ikl, c1b+ikl+1):
                if ci < 0 or ci >= as1:
                    continue
                cwi = mks + (ci-c1b) * ks - int(c1o * fks)
                if cwi < 0 or cwi >= nk:
                    continue
                kwi = kwi0 * k[cwi]
                valp = a[ri,ci]
                if not numpy.isnan(valp):
                    val += kwi * valp
                    vw += kwi
        if vw == 0.0:
            vw = 1.0
        out[i] = val / vw
    return out

# sample grid coordinates
@jit([
    'f8[:](f8[:,:],f8[:,:],f8[:],i8)', #output(array, crd, kernel, ksize)
    'f8[:](f4[:,:],f8[:,:],f8[:],i8)',
    'f8[:](u1[:,:],f8[:,:],f8[:],i8)',
    ], nopython=True)
def _sample_grid_coords_fine(
    a:numpy.ndarray,
    c:numpy.ndarray,
    k:numpy.ndarray,
    ks:int64) -> numpy.ndarray:
    nk = k.size -1
    kl = float(nk) / float(2 * ks)
    if kl != numpy.trunc(kl):
        raise ValueError('Invalid kernel.')
    ikl = int(kl)
    mks = ikl * ks
    fks = float(ks)
    as0 = a.shape[0]
    as1 = a.shape[1]
    nc = c.shape[0]
    if c.shape[1] != 2:
        raise ValueError('Invalid coordinate list.')
    out = numpy.zeros(nc, dtype=numpy.float64).reshape((nc,))
    for i in prange(nc): #pylint: disable=not-an-iterable
        c0c = c[i,0]
        c0b = int(c0c + 0.5)
        c0o = c0c - float(c0b)
        c1c = c[i,1]
        c1b = int(c1c + 0.5)
        c1o = c1c - float(c1b)
        val = 0.0
        vw = 0.0
        for ri in range(c0b-ikl, c0b+ikl+1):
            if ri < 0 or ri >= as0:
                continue
            wf = c0o * fks
            wfi = int(wf)
            wfp = wf - float(wfi)
            wi = mks + (ri-c0b) * ks - wfi
            if wi <= 0 or wi >= nk:
                continue
            kwi0 = (1.0 - wfp) * k[wi] + wfp * k[wi-1]
            for ci in range(c1b-ikl, c1b+ikl+1):
                if ci < 0 or ci >= as1:
                    continue
                wf = c1o * fks
                wfi = int(wf)
                wfp = wf - float(wfi)
                cwi = mks + (ci-c1b) * ks - wfi
                if cwi <= 0 or cwi >= nk:
                    continue
                kwi = kwi0 * ((1.0 - wfp) * k[cwi] + wfp * k[cwi-1])
                valp = a[ri,ci]
                if not numpy.isnan(valp):
                    val += kwi * valp
                    vw += kwi
        if vw == 0.0:
            vw = 1.0
        out[i] = val / vw
    return out

# sample values
@jit([
    'f8[:](f8[:],f8[:],f8[:],i8)', #output(vector, crd0, kernel, ksize)
    'f8[:](f4[:],f8[:],f8[:],i8)',
    'f8[:](u1[:],f8[:],f8[:],i8)',
    ], nopython=True)
def _sample_values(
    a:numpy.ndarray,
    c:numpy.ndarray,
    k:numpy.ndarray,
    ks:int64) -> numpy.ndarray:
    nk = k.size - 1
    kl = float(nk) / float(2 * ks)
    if kl != numpy.trunc(kl):
        raise ValueError('Invalid kernel.')
    ikl = int(kl)
    mks = ikl * ks
    fks = float(ks)
    nc = c.size
    al = a.size
    v = numpy.zeros(nc, dtype=numpy.float64)
    for i0 in prange(nc): #pylint: disable=not-an-iterable
            c0c = c[i0]
            c0b = int(c0c + 0.5)
            c0o = c0c - float(c0b)
            val = 0.0
            vw = 0.0
            for ci in range(c0b-ikl, c0b+ikl+1):
                if ci < 0 or ci >= al:
                    continue
                cwi = mks + (ci-c0b) * ks - int(c0o * fks)
                if cwi > 0 and cwi < nk:
                    kwi = k[cwi]
                    valp = a[ci]
                    if not numpy.isnan(valp):
                        val += kwi * valp
                        vw += kwi
            if vw == 0.0:
                vw = 1.0
            v[i0] = val / vw
    return v

def trans_matrix(m:Union[list, dict, tuple]) -> numpy.ndarray:
    if isinstance(m, tuple):
        mt = m
        m = dict()
        if len(mt) > 0:
            m['trans'] = mt[0]
        if len(mt) > 1:
            m['rotate'] = mt[1]
        if len(mt) > 2:
            m['scale'] = mt[2]
        if len(mt) > 3:
            m['shear'] = mt[3]
        if len(mt) > 4:
            m['origin'] = mt[4]
    if isinstance(m, dict):
        m = [m]
    elif not isinstance(m, list):
        raise ValueError('Invalid input parameter.')
    elif len(m) < 1:
        raise ValueError('Invalid input parameter.')
    nd = 0
    try:
        for m_in in m:
            if not isinstance(m_in, dict):
                raise ValueError('Invalid input parameter.')
            origin = m_in.get('origin', None)
            if not origin is None:
                lt = len(origin)
                if nd > 0 and nd != lt:
                    raise ValueError('Invalid origin field in input parameter.')
                elif nd == 0:
                    nd = lt
            else:
                if nd == 0:
                    m_in['origin'] = None
                else:
                    m_in['origin'] = numpy.zeros(nd, numpy.float64)
            trans = m_in.get('trans', None)
            if not trans is None:
                lt = len(trans)
                if nd > 0 and nd != lt:
                    raise ValueError('Invalid origin field in input parameter.')
                elif nd == 0:
                    nd = lt
            else:
                if nd == 0:
                    m_in['trans'] = None
                else:
                    m_in['trans'] = numpy.zeros(nd, numpy.float64)
            rotate = m_in.get('rotate', None)
            if not rotate is None:
                lt = len(rotate)
                if not lt in [1, 3]:
                    raise ValueError('Invalid rotate field in input parameter.')
                elif lt == 1:
                    lt = 2
                if nd > 0 and nd != lt:
                    raise ValueError('Invalid rotate field in input parameter.')
                elif nd == 0:
                    nd = lt
            else:
                if nd == 2:
                    m_in['rotate'] = numpy.zeros(1, numpy.float64)
                elif nd == 3:
                    m_in['rotate'] = numpy.zeros(3, numpy.float64)
                else:
                    m_in['rotate'] = None
            scale = m_in.get('scale', None)
            if not scale is None:
                lt = len(scale)
                if lt > 1:
                    if nd > 0 and nd != lt:
                        raise ValueError('Invalid scale field in input parameter.')
                    elif nd == 0:
                        nd = lt
                elif nd > 0:
                        m_in['scale'] = scale * numpy.ones(nd, numpy.float64)
            else:
                if nd == 0:
                    m_in['scale'] = None
                else:
                    m_in['scale'] = numpy.ones(nd, numpy.float64)
            shear = m_in.get('shear', None)
            if not shear is None:
                lt = len(shear)
                if not lt in [1, 3]:
                    raise ValueError('Invalid shear field in input parameter.')
                elif lt == 1:
                    lt = 2
                if nd > 0 and nd != lt:
                    raise ValueError('Invalid rotate field in input parameter.')
                elif nd == 0:
                    nd = lt
            else:
                if nd == 2:
                    m_in['shear'] = numpy.zeros(1, numpy.float64)
                elif nd == 3:
                    m_in['shear'] = numpy.zeros(3, numpy.float64)
                else:
                    m_in['shear'] = None
        if not nd in [2,3]:
            raise ValueError('Invalid input parameter (dimensions not inferred).')
        m_out = numpy.zeros((nd+1, nd+1,), numpy.float64)
        for n in range(nd+1):
            m_out[n,n] = 1.0
        m_p = m_out.copy()
        for m_in in reversed(m):
            origin = m_in['origin']
            if origin is None:
                origin = numpy.zeros(nd, numpy.float64)
            trans = m_in['trans']
            if trans is None:
                trans = numpy.zeros(nd, numpy.float64)
            rotate = m_in['rotate']
            if rotate is None:
                if nd == 2:
                    rotate = numpy.zeros(1, numpy.float64)
                else:
                    rotate = numpy.zeros(3, numpy.float64)
            rs = numpy.sin(rotate)
            rc = numpy.cos(rotate)
            scale = m_in['scale']
            if scale is None:
                scale = numpy.ones(nd, numpy.float64)
            shear = m_in['shear']
            if shear is None:
                if nd == 2:
                    shear = numpy.zeros(1, numpy.float64)
                else:
                    shear = numpy.zeros(3, numpy.float64)
            m_o = m_p.copy()
            m_o[:nd,-1] = origin
            m_ob = m_p.copy()
            m_ob[:nd,-1] = -origin
            m_t = m_p.copy()
            m_t[:nd,-1] = trans
            if nd == 2:
                m_r = m_p.copy()
                m_r[0:2,0:2] = numpy.asarray([[rc[0], rs[0]], [-rs[0], rc[0]]])
            else:
                m_r1 = m_p.copy()
                m_r1[1:3,1:3] = numpy.asarray([[rc[0], rs[0]], [-rs[0], rc[0]]])
                m_r2 = m_p.copy()
                m_r2[0,0] = rc[1]
                m_r2[0,2] = rs[1]
                m_r2[2,0] = -rs[1]
                m_r2[2,2] = rc[1]
                m_r3 = m_p.copy()
                m_r3[0:2,0:2] = numpy.asarray([[rc[2], rs[2]], [-rs[2], rc[2]]])
                m_r = numpy.matmul(numpy.matmul(m_r1, m_r2), m_r3)
            m_s = m_p.copy()
            for n in range(nd):
                m_s[n,n] = scale[n]
            m_h = m_p.copy()
            m_h[0,1] = shear[0]
            if nd == 3:
                m_h[0,2] = shear[1]
                m_h[1,2] = shear[2]
            m_c = numpy.matmul(
                m_t, numpy.matmul(
                m_o, numpy.matmul(
                m_r, numpy.matmul(
                m_s, numpy.matmul(
                m_ob,
                m_h)))))
            m_out = numpy.matmul(m_c, m_out)
    except:
        raise
    return m_out

def trans_matrix_inv(m:numpy.ndarray):
    """
    Decompose transformation matrix into parts

    Parameters
    ----------
    m : ndarray
        Transformation matrix
    
    Returns
    -------
    trans : ndarray
        2- or 3-element translation
    rotate : ndarray
        1- or 3-element rotation angles

    """
    was2d = False
    if m.shape[1] == 3:
        was2d = True
        m = numpy.asarray([
            [1.0, 0.0, 0.0, 0.0],
            [0.0, m[0,0], m[0,1], m[0,2]],
            [0.0, m[1,0], m[1,1], m[1,2]],
            [0.0, 0.0, 0.0, 1.0]], numpy.float64)
    trans = m[0:3,3]
    rotate = numpy.zeros(3, numpy.float64)
    r = m[0:3,0:3]
    rc = numpy.linalg.cholesky(numpy.matmul(r.T, r)).T
    scale = numpy.diagonal(rc)
    if numpy.linalg.det(r) < 0.0:
        scale[0] *= -1.0
    rcd = rc * numpy.eye(3, dtype=numpy.float64)
    rc = numpy.linalg.solve(rcd, rc)
    shear = numpy.asarray([rc[0,1], rc[0,2], rc[1,2]], numpy.float64)
    r0 = trans_matrix({'rotate': rotate, 'scale': scale, 'shear': shear})[0:3,0:3]
    r0 = numpy.linalg.solve(numpy.linalg.inv(r), numpy.linalg.inv(r0))
    rotate[1] = numpy.arcsin(_frone(r0[0,2]))
    if numpy.abs((numpy.abs(rotate[1]) - (numpy.pi / 2.0))) < 1.0e-6:
        rotate[0] = 0.0
        rotate[2] = numpy.arctan2(-_frone(r0[1,0]), _frone(-r0[2,0] / r0[0,2]))
    else:
        rc = numpy.cos(rotate[1])
        rotate[0] = numpy.arctan2(_frone(r0[1,2] / rc), _frone(r0[2,2] / rc))
        rotate[2] = numpy.arctan2(_frone(r0[0,1] / rc), _frone(r0[0,0] / rc))
    if was2d:
        trans = trans[1:]
        rotate = rotate[0:1]
        scale = scale[1:]
        shear = shear[2:3]
    return (trans, rotate, scale, shear)


class Sampler(object):


    __kernels = {}
    def __init__(self):

        self._kernels = self.__kernels

        # prepare some kernels
        if not self._kernels:
            ks = 4096
            kn = numpy.zeros(2*ks+1, dtype=numpy.float64)
            kn[ks//2+1:ks+ks//2] = 1.0
            kn[ks//2] = 0.5
            kn[ks+ks//2] = 0.5
            self._kernels['nearest'] = [kn, ks]
            kn = numpy.zeros(2*ks+1, dtype=numpy.float64)
            kn[0:ks] = numpy.arange(0.0, 1.0, 1.0 / float(ks))
            kn[ks:2*ks] = numpy.arange(1.0, 0.0, -1.0 / float(ks))
            self._kernels['linear'] = [kn, ks]
            k21 = [v for v in range(0,ks)]
            k22 = [v for v in range(ks+1,3*ks)]
            k23 = [v for v in range(3*ks+1,4*ks)]
            k = numpy.abs(numpy.arange(-2.0, 2.0+0.5/float(ks), 1.0/float(ks)).astype(numpy.float64))
            k[k21] = -0.5 * (k[k21] ** 3) + 2.5 * (k[k21] * k[k21]) - 4.0 * k[k21] + 2.0
            k[k22] = 1.5 * (k[k22] ** 3) - 2.5 * (k[k22] * k[k22]) + 1.0
            k[k23] = -0.5 * (k[k23] ** 3) + 2.5 * (k[k23] * k[k23]) - 4.0 * k[k23] + 2.0
            k[0::ks] = 0.0
            k[2*ks] = 1.0
            self._kernels['cubic'] = [k, ks]
            kss = [0, 0, 8192, 8192, 4096, 4096, 2048, 2048, 2048, 2048]
            math_pi = numpy.float(3.1415926535897931)
            for kc in range(2,10):
                ks = kss[kc]
                k = numpy.arange(-float(kc), float(kc) + 0.5/float(ks), 1.0/float(ks)).astype(numpy.float64)
                k[kc*ks] = 1.0
                pi_k = math_pi * k
                ksin = numpy.sin(math_pi * k) / (pi_k * pi_k)
                ksin[0::ks] = 0.0
                ksin = (kc * ksin) * numpy.sin((math_pi / float(kc)) * k)
                ksin[kc*ks] = 1.0
                self._kernels['lanczos' + str(kc)] = [ksin, ks]
    
    # sample values
    def sample_values(self,
        a:numpy.ndarray,
        s:Union[numpy.ndarray,list,tuple,int,float],
        k:Union[str,tuple] = 'resample',
        out_type:str = 'float64',
        ) -> numpy.ndarray:
        if not isinstance(a, numpy.ndarray):
            raise ValueError('Invalid array a to sample.')
        ad = a.ndim
        ash = a.shape
        if isinstance(s, int):
            s = float(s) / float(ash[0])
        if isinstance(s, float):
            s = [int(s * (float(ash[0]) + 0.5))]
        if isinstance(s, numpy.ndarray):
            if s.ndim != 1 or s.shape[0] != 3:
                raise ValueError('Invalid sampling specification.')
            s = numpy.arange(s[0], s[1], s[2]).astype(numpy.float64)
        elif (not isinstance(s, list) and not isinstance(s, tuple)) or len(s) > 1:
            raise ValueError('Invalid sampling specification.')
        else:
            s = s[0]
            try:
                if isinstance(s, int):
                    sf = float(ash[0]) / float(s)
                    s = numpy.arange(sf/2.0-0.5, float(ash[0])-0.5, sf)
                elif not isinstance(s, numpy.ndarray):
                    s = numpy.asarray(s).astype(numpy.float64)
            except:
                raise
        if isinstance(k, str):
            if k == 'resample':
                fs = numpy.mean(numpy.diff(s))
                fm = 0.1 * numpy.trunc(10.0 * fs)
                if fm <= 1.0:
                    k = self._kernels['cubic']
                else:
                    fms = 'rs_{0:.1f}'.format(fm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        kc = self._kernels['cubic']
                        sk = _gauss_kernel(fm * float(kc[1])).astype(numpy.float64)
                        skl = sk.size
                        skr = (skl - 1) // (2 * kc[1])
                        skr = 2 * kc[1] * skr + 1
                        skd = (skl - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        ksk = numpy.convolve(kc[0], sk)
                        while numpy.sum(ksk[0:kc[1]]) < 0.01:
                            ksk = ksk[kc[1]:-kc[1]]
                        k = [ksk, kc[1]]
                        self._kernels[fms] = k
            elif len(k) > 5 and k[0:5] == 'gauss':
                try:
                    fwhm = 0.1 * float(int(0.5 + 10 * float(k[5:])))
                    fms = 'g_{0:.1f}'.format(fwhm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        sk = _gauss_kernel(fwhm * float(1024))
                        skr = (sk.size - 1) // 2048
                        skr = 2048 * skr + 1
                        skd = (sk.size - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        k = [sk, 1024]
                        self._kernels[fms] = k
                except:
                    raise ValueError('Invalid gaussian kernel requested.')
            elif not k in self._kernels:
                raise ValueError('Kernel ' + k + ' not available.')
            else:
                k = self._kernels[k]
        elif not isinstance(k, tuple) or len(k) != 2 or (
            not isinstance(k[0], numpy.ndarray) or len(k[1]) != 1 or
            (float(k[0].size - 1) % float(k[1])) != 0.0):
            raise ValueError('Invalid kernel k.')
        if ad == 1:
            out = _sample_values(a, s, k[0], k[1])
        else:
            as0 = ash[0]
            if ad > 3:
                raise ValueError('Invalid data provided for column-based sampling.')
            out = _sample_values(a[:,0].reshape(as0), s, k[0], k[1])
            out.shape = (out.size,1,)
            for d in range(1, ad):
                out = numpy.repeat(out, ash[d], axis=d)
            if ad == 2:
                for d in range(1, ash[1]):
                    out[:,d] = _sample_values(a[:,d].reshape(as0), s, k[0], k[1])
            else:
                for d1 in range(ash[1]):
                    for d2 in range(ash[2]):
                        out[:,d1,d2] = _sample_values(a[:,d1,d2].reshape(as0), s, k[0],k[1])
        if out_type != 'float64':
            if out_type == 'uint8':
                out = numpy.minimum(numpy.maximum(out, 0.0), 255.0).astype(numpy.uint8)
            elif out_type == 'float32':
                out = out.astype(numpy.float32)
            elif out_type == 'int16':
                out = numpy.minimum(numpy.maximum(out, -32768.0), 32767.0).astype(numpy.int16)
            elif out_type == 'int32':
                out = out.astype(numpy.int32)
            else:
                warnings.warn('Output of type ' + out_type + ' not supported; returning float64.')
        return out

    # sample 2D grid
    def sample_grid(self,
        a:numpy.ndarray,
        s:Union[numpy.ndarray,list,tuple,int,float],
        k:Union[str,tuple] = 'resample',
        out_type:str = 'float64',
        m:Union[list,dict,numpy.ndarray] = None,
        fine:bool = False,
        ) -> numpy.ndarray:
        if not isinstance(a, numpy.ndarray):
            raise ValueError('Invalid array a to sample.')
        ad = a.ndim
        ash = a.shape
        if isinstance(s, int):
            s = float(s) / float(max(ash[0], ash[1]))
        if isinstance(s, float):
            sf = s
            s = []
            for d in range(min(2,ad)):
                s.append(int(sf * (float(ash[d]) + 0.5)))
        if isinstance(s, numpy.ndarray):
            if s.ndim != 2 or s.shape[0] != 3 or s.shape[1] > ad:
                raise ValueError('Invalid sampling specification.')
            sl = []
            for d in s.shape[1]:
                sl.append(numpy.arange(s[0,d], s[1,d], s[2,d]).astype(numpy.float64))
            s = sl
        elif (not isinstance(s, list) and not isinstance(s, tuple)) or len(s) > ad:
            raise ValueError('Invalid sampling specification.')
        else:
            s = s[:]
            try:
                for d in range(len(s)):
                    if isinstance(s[d], int):
                        sf = float(ash[d]) / float(s[d])
                        s[d] = numpy.arange(sf/2.0-0.5, float(ash[d])-0.5, sf)
                    elif isinstance(s[d], float):
                        sf = 1.0 / s[d]
                        s[d] = numpy.arange(sf/2.0-0.5, float(ash[d])-0.5, sf)
                    elif not isinstance(s[d], numpy.ndarray):
                        s[d] = numpy.asarray(s[d]).astype(numpy.float64)
            except:
                raise
        if isinstance(k, str):
            if k == 'resample':
                fs = []
                for d in range(len(s)):
                    fs.append(numpy.mean(numpy.diff(s[d])))
                fm = 0.1 * numpy.trunc(10.0 * numpy.mean(fs))
                if fm <= 1.0:
                    k = self._kernels['cubic']
                else:
                    fms = 'rs_{0:.1f}'.format(fm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        kc = self._kernels['cubic']
                        sk = _gauss_kernel(fm * float(kc[1])).astype(numpy.float64)
                        skl = sk.size
                        skr = (skl - 1) // (2 * kc[1])
                        skr = 2 * kc[1] * skr + 1
                        skd = (skl - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        ksk = numpy.convolve(kc[0], sk)
                        while numpy.sum(ksk[0:kc[1]]) < 0.01:
                            ksk = ksk[kc[1]:-kc[1]]
                        k = [ksk, kc[1]]
                        self._kernels[fms] = k
            elif len(k) > 5 and k[0:5] == 'gauss':
                try:
                    fwhm = 0.1 * float(int(0.5 + 10 * float(k[5:])))
                    fms = 'g_{0:.1f}'.format(fwhm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        sk = _gauss_kernel(fwhm * float(1024))
                        skr = (sk.size - 1) // 2048
                        skr = 2048 * skr + 1
                        skd = (sk.size - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        k = [sk, 1024]
                        self._kernels[fms] = k
                except:
                    raise ValueError('Invalid gaussian kernel requested.')
            elif not k in self._kernels:
                raise ValueError('Kernel ' + k + ' not available.')
            else:
                k = self._kernels[k]
        elif not isinstance(k, tuple) or len(k) != 2 or (
            not isinstance(k[0], numpy.ndarray) or len(k[1]) != 1 or
            (float(k[0].size - 1) % float(k[1])) != 0.0):
            raise ValueError('Invalid kernel k.')
        ls = len(s)
        if ls == 2:
            if isinstance(m, list) or isinstance(m, dict):
                try:
                    m = trans_matrix(m)
                except:
                    raise
            if m is None or not isinstance(m, numpy.ndarray):
                if ad == 2:
                    out = _sample_grid_2d(a, s[0], s[1], k[0], k[1])
                elif ad == 3:
                    out = _sample_grid_2d(a[:,:,0].reshape((ash[0], ash[1],)),
                        s[0], s[1], k[0], k[1])
                    outsh = out.shape
                    out = numpy.repeat(out.reshape((outsh[0], outsh[1], 1)),
                        ash[2], axis=2)
                    for p in range(1, ash[2]):
                        out[:,:,p] = _sample_grid_2d(a[:,:,p].reshape((ash[0], ash[1])),
                            s[0], s[1], k[0], k[1]).reshape((outsh[0], outsh[1],))
                else:
                    raise ValueError('Sampling 2D grid of 4D data not supported.')
            else:
                if m.dtype != numpy.float64 or m.shape[1] != 3 or m.shape[0] < 2:
                    raise ValueError('Invalid transformation matrix m.')
                s0 = s[0].size
                s1 = s[1].size
                (c1, c0) = numpy.meshgrid(s[1], s[0])
                c0.shape = (c0.size,1,)
                c1.shape = (c1.size,1,)
                c01 = numpy.concatenate(
                    (m[0,0]*c0+m[0,1]*c1+m[0,2], m[1,0]*c0+m[1,1]*c1+m[1,2]), axis=1)
                if ad == 2:
                    if fine:
                        out = _sample_grid_coords_fine(
                            a, c01, k[0], k[1]).reshape((s0,s1,))
                    else:
                        out = _sample_grid_coords(
                            a, c01, k[0], k[1]).reshape((s0,s1,))
                elif ad == 3:
                    outsh = (s0,s1,1,)
                    if fine:
                        out = _sample_grid_coords_fine(
                            a[:,:,0].reshape((ash[0], ash[1],)),
                            c01, k[0], k[1]).reshape(outsh)
                        out = numpy.repeat(out, ash[2], axis=2)
                        for p in range(1, ash[2]):
                            out[:,:,p] = _sample_grid_coords_fine(
                                a[:,:,p].reshape((ash[0], ash[1],)),
                                c01, k[0], k[1]).reshape((s0,s1,))
                    else:
                        out = _sample_grid_coords(
                            a[:,:,0].reshape((ash[0], ash[1],)),
                            c01, k[0], k[1]).reshape(outsh)
                        out = numpy.repeat(out, ash[2], axis=2)
                        for p in range(1, ash[2]):
                            out[:,:,p] = _sample_grid_coords(
                                a[:,:,p].reshape((ash[0], ash[1],)),
                                c01, k[0], k[1]).reshape((s0,s1,))
        elif ls == 1:
            out = _sample_values(a, s[0], k[0], k[1])
        elif ls == 3:
            raise NotImplementedError('3D interpolation not yet implemented.')
        else:
            raise NotImplementedError('Higher dim interpolation not yet implemented.')
        if out_type != 'float64':
            if out_type == 'uint8':
                out = numpy.minimum(numpy.maximum(out, 0.0), 255.0).astype(numpy.uint8)
            elif out_type == 'float32':
                out = out.astype(numpy.float32)
            elif out_type == 'int16':
                out = numpy.minimum(numpy.maximum(out, -32768.0), 32767.0).astype(numpy.int16)
            elif out_type == 'int32':
                out = out.astype(numpy.int32)
            else:
                warnings.warn('Output of type ' + out_type + ' not supported; returning float64.')
        return out

    # sample 2D grid
    def sample_radial(self,
        a:numpy.ndarray,
        cx:float,
        cy:float,
        step:float,
        steps:int,
        astep:float,
        k:Union[str,tuple] = 'cubic',
        out_type:str = 'float64',
        fine:bool = False,
        ) -> numpy.ndarray:
        if not isinstance(a, numpy.ndarray):
            raise ValueError('Invalid array a to sample.')
        ad = a.ndim
        ash = a.shape
        if not isinstance(cx, float) or not isinstance(cy, float):
            raise ValueError('Invalid center coordinate.')
        if not isinstance(step, float) or step <= 0.0:
            raise ValueError('Invalid step size.')
        if not isinstance(steps, int) or steps <= 0:
            raise ValueError('Invalid number of steps.')
        steps
        if not isinstance(astep, float) or astep <= 0.0:
            raise ValueError('Invalid angular step size.')
        astep = numpy.arange(0.0, numpy.pi/2.0, astep)
        if isinstance(s, int):
            s = float(s) / float(max(ash[0], ash[1]))
        if isinstance(s, float):
            sf = s
            s = []
            for d in range(min(2,ad)):
                s.append(int(sf * (float(ash[d]) + 0.5)))
        if isinstance(s, numpy.ndarray):
            if s.ndim != 2 or s.shape[0] != 3 or s.shape[1] > ad:
                raise ValueError('Invalid sampling specification.')
            sl = []
            for d in s.shape[1]:
                sl.append(numpy.arange(s[0,d], s[1,d], s[2,d]).astype(numpy.float64))
            s = sl
        elif (not isinstance(s, list) and not isinstance(s, tuple)) or len(s) > ad:
            raise ValueError('Invalid sampling specification.')
        else:
            s = s[:]
            try:
                for d in range(len(s)):
                    if isinstance(s[d], int):
                        sf = float(ash[d]) / float(s[d])
                        s[d] = numpy.arange(sf/2.0-0.5, float(ash[d])-0.5, sf)
                    elif isinstance(s[d], float):
                        sf = 1.0 / s[d]
                        s[d] = numpy.arange(sf/2.0-0.5, float(ash[d])-0.5, sf)
                    elif not isinstance(s[d], numpy.ndarray):
                        s[d] = numpy.asarray(s[d]).astype(numpy.float64)
            except:
                raise
        if isinstance(k, str):
            if k == 'resample':
                fs = []
                for d in range(len(s)):
                    fs.append(numpy.mean(numpy.diff(s[d])))
                fm = 0.1 * numpy.trunc(10.0 * numpy.mean(fs))
                if fm <= 1.0:
                    k = self._kernels['cubic']
                else:
                    fms = 'rs_{0:.1f}'.format(fm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        kc = self._kernels['cubic']
                        sk = _gauss_kernel(fm * float(kc[1])).astype(numpy.float64)
                        skl = sk.size
                        skr = (skl - 1) // (2 * kc[1])
                        skr = 2 * kc[1] * skr + 1
                        skd = (skl - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        ksk = numpy.convolve(kc[0], sk)
                        while numpy.sum(ksk[0:kc[1]]) < 0.01:
                            ksk = ksk[kc[1]:-kc[1]]
                        k = [ksk, kc[1]]
                        self._kernels[fms] = k
            elif len(k) > 5 and k[0:5] == 'gauss':
                try:
                    fwhm = 0.1 * float(int(0.5 + 10 * float(k[5:])))
                    fms = 'g_{0:.1f}'.format(fwhm)
                    if fms in self._kernels:
                        k = self._kernels[fms]
                    else:
                        sk = _gauss_kernel(fwhm * float(1024))
                        skr = (sk.size - 1) // 2048
                        skr = 2048 * skr + 1
                        skd = (sk.size - skr) // 2
                        sk = sk[skd:skr+skd]
                        sk = sk / numpy.sum(sk)
                        k = [sk, 1024]
                        self._kernels[fms] = k
                except:
                    raise ValueError('Invalid gaussian kernel requested.')
            elif not k in self._kernels:
                raise ValueError('Kernel ' + k + ' not available.')
            else:
                k = self._kernels[k]
        elif not isinstance(k, tuple) or len(k) != 2 or (
            not isinstance(k[0], numpy.ndarray) or len(k[1]) != 1 or
            (float(k[0].size - 1) % float(k[1])) != 0.0):
            raise ValueError('Invalid kernel k.')
        ls = len(s)
        if ls == 2:
            if isinstance(m, list) or isinstance(m, dict):
                try:
                    m = trans_matrix(m)
                except:
                    raise
            if m is None or not isinstance(m, numpy.ndarray):
                if ad == 2:
                    out = _sample_grid_2d(a, s[0], s[1], k[0], k[1])
                elif ad == 3:
                    out = _sample_grid_2d(a[:,:,0].reshape((ash[0], ash[1],)),
                        s[0], s[1], k[0], k[1])
                    outsh = out.shape
                    out = numpy.repeat(out.reshape((outsh[0], outsh[1], 1)),
                        ash[2], axis=2)
                    for p in range(1, ash[2]):
                        out[:,:,p] = _sample_grid_2d(a[:,:,p].reshape((ash[0], ash[1])),
                            s[0], s[1], k[0], k[1]).reshape((outsh[0], outsh[1],))
                else:
                    raise ValueError('Sampling 2D grid of 4D data not supported.')
            else:
                if m.dtype != numpy.float64 or m.shape[1] != 3 or m.shape[0] < 2:
                    raise ValueError('Invalid transformation matrix m.')
                s0 = s[0].size
                s1 = s[1].size
                (c1, c0) = numpy.meshgrid(s[1], s[0])
                c0.shape = (c0.size,1,)
                c1.shape = (c1.size,1,)
                c01 = numpy.concatenate(
                    (m[0,0]*c0+m[0,1]*c1+m[0,2], m[1,0]*c0+m[1,1]*c1+m[1,2]), axis=1)
                if ad == 2:
                    if fine:
                        out = _sample_grid_coords_fine(
                            a, c01, k[0], k[1]).reshape((s0,s1,))
                    else:
                        out = _sample_grid_coords(
                            a, c01, k[0], k[1]).reshape((s0,s1,))
                elif ad == 3:
                    outsh = (s0,s1,1,)
                    if fine:
                        out = _sample_grid_coords_fine(
                            a[:,:,0].reshape((ash[0], ash[1],)),
                            c01, k[0], k[1]).reshape(outsh)
                        out = numpy.repeat(out, ash[2], axis=2)
                        for p in range(1, ash[2]):
                            out[:,:,p] = _sample_grid_coords_fine(
                                a[:,:,p].reshape((ash[0], ash[1],)),
                                c01, k[0], k[1]).reshape((s0,s1,))
                    else:
                        out = _sample_grid_coords(
                            a[:,:,0].reshape((ash[0], ash[1],)),
                            c01, k[0], k[1]).reshape(outsh)
                        out = numpy.repeat(out, ash[2], axis=2)
                        for p in range(1, ash[2]):
                            out[:,:,p] = _sample_grid_coords(
                                a[:,:,p].reshape((ash[0], ash[1],)),
                                c01, k[0], k[1]).reshape((s0,s1,))
        elif ls == 1:
            out = _sample_values(a, s[0], k[0], k[1])
        elif ls == 3:
            raise NotImplementedError('3D interpolation not yet implemented.')
        else:
            raise NotImplementedError('Higher dim interpolation not yet implemented.')
        if out_type != 'float64':
            if out_type == 'uint8':
                out = numpy.minimum(numpy.maximum(out, 0.0), 255.0).astype(numpy.uint8)
            elif out_type == 'float32':
                out = out.astype(numpy.float32)
            elif out_type == 'int16':
                out = numpy.minimum(numpy.maximum(out, -32768.0), 32767.0).astype(numpy.int16)
            elif out_type == 'int32':
                out = out.astype(numpy.int32)
            else:
                warnings.warn('Output of type ' + out_type + ' not supported; returning float64.')
        return out



# isicarchive package
"""
ISIC Archive API

Provides a python interface to https://isic-archive.com/api/v1

To instantiate the API object and retrieve some study information use

   >>> from isicarchive.api import IsicApi
   >>> api = IsicApi(username)
   >>> study_info = api.study(study_name)

Please consult the help of the IsicApi object class for further
documentation.

:copyright: (c) 2019, Jochen Weber, MSKCC.
:license: MIT, see LICENSE for details.
"""

# all previous imports removed; wasting time during import

from .version import __version__

name = 'isicarchive'



# from https://github.com/ImageMarkup/isic-archive/blob/master/isic_archive/models/masterFeatures.json

# order
# - circles / semicircles (24++)
# - dots (73++)
# - facial skin (163++)
# - globules / clods (215++)
#   - lacunae (260++)
# - lines (363++)
# - miscellaneous
# - nail lesions
# - network
# - pattern
#   - homogeneous pattern
# - regression structures
# - shiny white structures
# - structureless
#   - homogeneous pattern
# - vessels
# - volar lesions

master_features = [
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : NOS",
        "abbreviation": "Circ:NOS",
        "color": [120,120,120],
        "synonyms": ["Circles & Semicircles : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : Asymmetric follicular openings",
        "abbreviation": "Circ:Asym.O.",
        "synonyms": [
            "Circles & Semicircles : Asymmetric follicular openings",
            "Facial Skin : Asymmetric follicular openings"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : Brown",
        "abbreviation": "Circ:Brown",
        "color": [184,136,112],
        "synonyms": [
            "Circles & Semicircles : Brown",
            "Facial Skin : Brown circles"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : Circle within a circle",
        "abbreviation": "Circ:C.w/iC.",
        "color": [160,128,120],
        "synonyms": [
            "Circles & Semicircles : Circle within a circle",
            "Facial Skin : Circle within a circle",
            "Miscellaneous : Circle within a circle"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : Fish scale (mucosal lesions)",
        "abbreviation": "Circ:FishSc.",
        "synonyms": ["Circles & Semicircles : Fish scale (mucosal lesions)"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : Gray",
        "abbreviation": "Circ:Gray",
        "synonyms": ["Circles & Semicircles : Gray"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Circles & Semicircles : White",
        "abbreviation": "Circ:White",
        "color": [192,192,192],
        "synonyms": ["Circles & Semicircles : White"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : NOS",
        "abbreviation": "Dots:NOS",
        "color": [48,48,48],
        "synonyms": ["Dots : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Annular-granular pattern",
        "abbreviation": "Dots:An-gran",
        "synonyms": [
            "Dots : Annular-granular pattern",
            "Facial Skin : Annular-granular pattern",
            "Miscellaneous : Annular-granular pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Black",
        "abbreviation": "Dots:Black",
        "color": [48,40,24],
        "synonyms": ["Dots : Black"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Blue-gray",
        "abbreviation": "Dots:Bl-gray",
        "color": [8,64,128],
        "synonyms": ["Dots : Blue-gray"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Brown",
        "abbreviation": "Dots:Brown",
        "color": [76,48,40],
        "synonyms": ["Dots : Brown"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Granularity",
        "abbreviation": "Dots:Granul.",
        "synonyms": ["Dots : Granularity"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Irregular",
        "abbreviation": "Dots:Irreg.",
        "color": [96,56,56],
        "synonyms": ["Dots : Irregular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Linear",
        "abbreviation": "Dots:Linear",
        "synonyms": ["Dots : Linear"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Milia-like cysts",
        "abbreviation": "Dots:MiliaC.",
        "synonyms": [
            "Dots : Milia-like cysts",
            "Globules / Clods : Milia-like cysts",
            "Miscellaneous : Milia-like cysts, cloudy or starry"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Peppering",
        "abbreviation": "Dots:Pepper.",
        "synonyms": [
            "Dots : Peppering",
            "Regression structures : Peppering / Granularity"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Regular",
        "abbreviation": "Dots:Regular",
        "color": [72,48,48],
        "synonyms": ["Dots : Regular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Rosettes",
        "abbreviation": "Dots:Rosette",
        "synonyms": [
            "Dots : Rosettes",
            "Shiny white structures : Rosettes"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Targetoid",
        "abbreviation": "Dots:Target.",
        "synonyms": ["Dots : Targetoid"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Dots : Yellow",
        "abbreviation": "Dots:Yellow",
        "synonyms": ["Dots : Yellow"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Annular-granular pattern",
        "abbreviation": "Face:An-gran",
        "synonyms": [
            "Dots : Annular-granular pattern",
            "Facial Skin : Annular-granular pattern",
            "Miscellaneous : Annular-granular pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Asymmetric follicular openings",
        "abbreviation": "Face:Asym.O.",
        "synonyms": [
            "Circles & Semicircles : Asymmetric follicular openings",
            "Facial Skin : Asymmetric follicular openings"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Brown circles",
        "abbreviation": "Face:BrownC.",
        "synonyms": [
            "Circles & Semicircles : Brown",
            "Facial Skin : Brown circles"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Circle within a circle",
        "abbreviation": "Face:C.w/iC.",
        "synonyms": [
            "Circles & Semicircles : Circle within a circle",
            "Facial Skin : Circle within a circle",
            "Miscellaneous : Circle within a circle"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Pseudonetwork",
        "abbreviation": "Face:PseuNet",
        "synonyms": [
            "Facial Skin : Pseudonetwork",
            "Miscellaneous : Pseudonetwork",
            "Structureless : Pseudonetwork"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Facial Skin : Rhomboids / Zig-zag pattern",
        "abbreviation": "Face:Rhb/Zig",
        "synonyms": [
            "Facial Skin : Rhomboids / Zig-zag pattern",
            "Lines : Rhomboids (facial skin)",
            "Miscellaneous : Rhomboids / Zig-zag pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : NOS",
        "abbreviation": "G/C.:NOS",
        "color": [128,96,48],
        "synonyms": ["Globules / Clods : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Blue",
        "abbreviation": "G/C.:Blue",
        "color": [16,48,96],
        "synonyms": ["Globules / Clods : Blue"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Blue-gray ovoid nests",
        "abbreviation": "G/C.:Bl-grON",
        "color": [32,56,112],
        "synonyms": ["Globules / Clods : Blue-gray ovoid nests"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Cobblestone pattern",
        "abbreviation": "G/C.:CobblS.",
        "color": [48,48,104],
        "synonyms": ["Globules / Clods : Cobblestone pattern"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Comedo-like openings",
        "abbreviation": "G/C.:ComedoO",
        "color": [192,136,112],
        "synonyms": [
            "Globules / Clods : Comedo-like openings",
            "Miscellaneous : Comedo-like openings"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Concentric",
        "abbreviation": "G/C.:Concent",
        "color": [184,120,88],
        "synonyms": ["Globules / Clods : Concentric"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Irregular",
        "abbreviation": "G/C.:Irreg.",
        "color": [144,64,120],
        "synonyms": ["Globules / Clods : Irregular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Lacunae : NOS",
        "abbreviation": "G/C.:LacuNOS",
        "synonyms": [
            "Globules / Clods : Lacunae : NOS",
            "Miscellaneous : Lacunae : NOS"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Lacunae : Black",
        "abbreviation": "G/C.:LacuBlk",
        "color": [56,56,16],
        "synonyms": [
            "Globules / Clods : Lacunae : Black",
            "Miscellaneous : Lacunae : Black"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Lacunae : Blue",
        "abbreviation": "G/C.:LacuBlu",
        "synonyms": [
            "Globules / Clods : Lacunae : Blue",
            "Miscellaneous : Lacunae : Blue"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Lacunae : Red",
        "abbreviation": "G/C.:LacuRed",
        "color": [144,24,16],
        "synonyms": [
            "Globules / Clods : Lacunae : Red",
            "Miscellaneous : Lacunae : Red"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Lacunae : Red-purple",
        "abbreviation": "G/C.:LacuR-P",
        "color": [144,16,80],
        "synonyms": [
            "Globules / Clods : Lacunae : Red-purple",
            "Miscellaneous : Lacunae : Red-purple"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Leaflike area",
        "abbreviation": "G/C.:LeaflkA",
        "color": [72,48,64],
        "synonyms": [
            "Globules / Clods : Leaflike area",
            "Lines : Leaf like areas",
            "Miscellaneous : Leaflike area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Milia-like cysts",
        "abbreviation": "G/C.:MiliaC.",
        "color": [224,208,192],
        "synonyms": [
            "Dots : Milia-like cysts",
            "Globules / Clods : Milia-like cysts",
            "Miscellaneous : Milia-like cysts, cloudy or starry"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Milky red",
        "abbreviation": "G/C.:MilkRed",
        "color": [240,160,120],
        "synonyms": ["Globules / Clods : Milky red"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Regular",
        "abbreviation": "G/C.:Regular",
        "color": [152,80,72],
        "synonyms": ["Globules / Clods : Regular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Rim of brown globules",
        "abbreviation": "G/C.:RimBrwn",
        "color": [184,112,96],
        "synonyms": ["Globules / Clods : Rim of brown globules"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Shiny white blotches and strands",
        "abbreviation": "G/C.:ShWhtB.",
        "color": [192,136,112],
        "synonyms": [
            "Globules / Clods : Shiny white blotches and strands",
            "Shiny white structures : Shiny white blotches and strands"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Spoke wheel areas",
        "abbreviation": "G/C.:SpkWhlA",
        "synonyms": [
            "Globules / Clods : Spoke wheel areas",
            "Lines : Spoke wheel areas",
            "Miscellaneous : Spoke wheel area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : Variant of spoke wheel area",
        "abbreviation": "G/C.:VarSpkW",
        "synonyms": ["Globules / Clods : Variant of spoke wheel area"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Globules / Clods : White",
        "abbreviation": "G/C.:White",
        "synonyms": ["Globules / Clods : White"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : NOS",
        "abbreviation": "Line:NOS",
        "color": [40,40,64],
        "synonyms": ["Lines : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Angulated lines / Polygons / Zig-zag pattern",
        "abbreviation": "Line:Ang/P/Z",
        "color": [96,32,32],
        "synonyms": [
            "Lines : Angulated lines / Polygons / Zig-zag pattern",
            "Miscellaneous : Angulated lines / Polygons"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Atypical pigment network / Reticulation",
        "abbreviation": "Line:AtypPNR",
        "synonyms": [
            "Lines : Atypical pigment network / Reticulation",
            "Network : Atypical pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Branched streaks",
        "abbreviation": "Line:BranSt.",
        "synonyms": ["Lines : Branched streaks"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Broadened pigment network / Reticulation",
        "abbreviation": "Line:Brd.PNR",
        "synonyms": [
            "Lines : Broadened pigment network / Reticulation",
            "Network : Broadened pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Cerebriform pattern",
        "abbreviation": "Line:CerebFP",
        "synonyms": [
            "Lines : Cerebriform pattern",
            "Pattern : Cerebriform"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Crypts",
        "abbreviation": "Line:Crypts",
        "synonyms": [
            "Lines : Crypts",
            "Miscellaneous : Crypts"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Delicate pigment network / Reticulation",
        "abbreviation": "Line:DeliPNR",
        "synonyms": [
            "Lines : Delicate pigment network / Reticulation",
            "Network : Delicate Pigment Network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Fibrillar pattern",
        "abbreviation": "Line:FibrilP",
        "synonyms": [
            "Lines : Fibrillar pattern",
            "Pattern : Fibrilar",
            "Volar lesions : Fibrilar pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Fingerprint pattern",
        "abbreviation": "Line:FingP.P",
        "synonyms": [
            "Lines : Fingerprint pattern",
            "Pattern : Fingerprint"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Leaf like areas",
        "abbreviation": "Line:LeaflkA",
        "synonyms": [
            "Globules / Clods : Leaflike area",
            "Lines : Leaf like areas",
            "Miscellaneous : Leaflike area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Negative pigment network",
        "abbreviation": "Line:Neg.PN.",
        "synonyms": [
            "Lines : Negative pigment network",
            "Network : Negative pigment network"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Parallel furrows pattern (volar lesions)",
        "abbreviation": "Line:ParFurP",
        "synonyms": [
            "Lines : Parallel furrows pattern (volar lesions)",
            "Pattern : Parallel furrow",
            "Volar lesions : Parallel furrow pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Parallel lines (volar lesions)",
        "abbreviation": "Line:ParLine",
        "synonyms": [
            "Lines : Parallel lines (volar lesions)",
            "Pattern : Parallel ridge",
            "Volar lesions : Parallel lines"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Parallel ridge pattern (volar lesions)",
        "abbreviation": "Line:ParRidP",
        "synonyms": [
            "Lines : Parallel ridge pattern (volar lesions)",
            "Volar lesions : Parallel ridge pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Pseudopods",
        "abbreviation": "Line:PseuPd.",
        "synonyms": [
            "Lines : Pseudopods",
            "Miscellaneous : Pseudopods"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Radial streaming",
        "abbreviation": "Line:RadStr.",
        "synonyms": [
            "Lines : Radial streaming",
            "Miscellaneous : Radial streaming"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Rhomboids (facial skin)",
        "abbreviation": "Line:RhbFac.",
        "synonyms": [
            "Facial Skin : Rhomboids / Zig-zag pattern",
            "Lines : Rhomboids (facial skin)",
            "Miscellaneous : Rhomboids / Zig-zag pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Shiny white streaks",
        "abbreviation": "Line:ShWhtS.",
        "synonyms": [
            "Lines : Shiny white streaks",
            "Shiny white structures : Shiny white streaks"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Spoke wheel areas",
        "abbreviation": "Line:SpkWhlA",
        "synonyms": [
            "Globules / Clods : Spoke wheel areas",
            "Lines : Spoke wheel areas",
            "Miscellaneous : Spoke wheel area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Starburst pattern",
        "abbreviation": "Line:StarbP.",
        "synonyms": [
            "Lines : Starburst pattern",
            "Pattern : Starburst"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Streaks",
        "abbreviation": "Line:Streaks",
        "synonyms": [
            "Lines : Streaks",
            "Miscellaneous : Streaks"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Lines : Typical pigment network / Reticulation",
        "abbreviation": "Line:Typ.PNR",
        "synonyms": [
            "Lines : Typical pigment network / Reticulation",
            "Network : Typical pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Angulated lines / Polygons",
        "abbreviation": "Misc:AngLine",
        "synonyms": [
            "Lines : Angulated lines / Polygons / Zig-zag pattern",
            "Miscellaneous : Angulated lines / Polygons"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Annular-granular pattern",
        "abbreviation": "Misc:An-gran",
        "synonyms": [
            "Dots : Annular-granular pattern",
            "Facial Skin : Annular-granular pattern",
            "Miscellaneous : Annular-granular pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Asymmetric pigmented follicular openings",
        "abbreviation": "Misc:Asym.O.",
        "synonyms": ["Miscellaneous : Asymmetric pigmented follicular openings"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Blood spots",
        "abbreviation": "Misc:BloodSp",
        "synonyms": [
            "Miscellaneous : Blood spots",
            "Nail lesions : Blood spots"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Blotch irregular",
        "abbreviation": "Misc:BlotIrr",
        "synonyms": [
            "Miscellaneous : Blotch irregular",
            "Structureless : Blotch irregular"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Blotch regular",
        "abbreviation": "Misc:BlotReg",
        "synonyms": [
            "Miscellaneous : Blotch regular",
            "Structureless : Blotch regular"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Blue-whitish veil",
        "abbreviation": "Misc:Blu-WhV",
        "synonyms": [
            "Miscellaneous : Blue-whitish veil",
            "Structureless : Blue-whitish veil"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Circle within a circle",
        "abbreviation": "Misc:C.w/iC.",
        "synonyms": [
            "Circles & Semicircles : Circle within a circle",
            "Facial Skin : Circle within a circle",
            "Miscellaneous : Circle within a circle"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Comedo-like openings",
        "abbreviation": "Misc:ComedoO",
        "synonyms": [
            "Globules / Clods : Comedo-like openings",
            "Miscellaneous : Comedo-like openings"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Crypts",
        "abbreviation": "Misc:Crypts",
        "synonyms": [
            "Lines : Crypts",
            "Miscellaneous : Crypts"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Disruption of parallelism",
        "abbreviation": "Misc:Disrup.",
        "synonyms": ["Miscellaneous : Disruption of parallelism"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Fissures",
        "abbreviation": "Misc:Fissure",
        "synonyms": ["Miscellaneous : Fissures"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Granularity or granules",
        "abbreviation": "Misc:Granul.",
        "synonyms": ["Miscellaneous : Granularity or granules"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Hutchinson sign",
        "abbreviation": "Misc:HutchS.",
        "synonyms": [
            "Miscellaneous : Hutchinson sign",
            "Nail lesions : Hutchinson sign"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Lacunae : NOS",
        "abbreviation": "Misc:LacuNOS",
        "synonyms": [
            "Globules / Clods : Lacunae : NOS",
            "Miscellaneous : Lacunae : NOS"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Lacunae : Black",
        "abbreviation": "Misc:LacuBlk",
        "synonyms": [
            "Globules / Clods : Lacunae : Black",
            "Miscellaneous : Lacunae : Black"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Lacunae : Blue",
        "abbreviation": "Misc:LacuBlu",
        "synonyms": [
            "Globules / Clods : Lacunae : Blue",
            "Miscellaneous : Lacunae : Blue"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Lacunae : Red",
        "abbreviation": "Misc:LacuRed",
        "color": [160,40,0],
        "synonyms": [
            "Globules / Clods : Lacunae : Red",
            "Miscellaneous : Lacunae : Red"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Lacunae : Red-purple",
        "abbreviation": "Misc:LacuR-P",
        "synonyms": [
            "Globules / Clods : Lacunae : Red-purple",
            "Miscellaneous : Lacunae : Red-purple"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Leaflike area",
        "abbreviation": "Misc:LacuLfA",
        "synonyms": [
            "Globules / Clods : Leaflike area",
            "Lines : Leaf like areas",
            "Miscellaneous : Leaflike area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Milia-like cysts, cloudy or starry",
        "abbreviation": "Misc:MiliaC.",
        "synonyms": [
            "Dots : Milia-like cysts",
            "Globules / Clods : Milia-like cysts",
            "Miscellaneous : Milia-like cysts, cloudy or starry"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Moth-eaten border",
        "abbreviation": "Misc:MothBor",
        "synonyms": ["Miscellaneous : Moth-eaten border"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Pseudo-Hutchinson sign",
        "abbreviation": "Misc:PseuHut",
        "synonyms": [
            "Miscellaneous : Pseudo-Hutchinson sign",
            "Nail lesions : Pseudo-Hutchinson sign"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Pseudonetwork",
        "abbreviation": "Misc:PseuNet",
        "synonyms": [
            "Facial Skin : Pseudonetwork",
            "Miscellaneous : Pseudonetwork",
            "Structureless : Pseudonetwork"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Pseudopods",
        "abbreviation": "Misc:PseuPod",
        "synonyms": [
            "Lines : Pseudopods",
            "Miscellaneous : Pseudopods"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Radial streaming",
        "abbreviation": "Misc:RadStr.",
        "synonyms": [
            "Lines : Radial streaming",
            "Miscellaneous : Radial streaming"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Regularly bended ribbon sign",
        "abbreviation": "Misc:RegBRib",
        "synonyms": ["Miscellaneous : Regularly bended ribbon sign"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Rhomboids / Zig-zag pattern",
        "abbreviation": "Misc:Rhb/Zig",
        "synonyms": [
            "Facial Skin : Rhomboids / Zig-zag pattern",
            "Lines : Rhomboids (facial skin)",
            "Miscellaneous : Rhomboids / Zig-zag pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Ridges",
        "abbreviation": "Misc:Ridges",
        "synonyms": ["Miscellaneous : Ridges"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Scale",
        "abbreviation": "Misc:Scale",
        "synonyms": ["Miscellaneous : Scale"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Setting-sun pattern",
        "abbreviation": "Misc:SunsetP",
        "synonyms": ["Miscellaneous : Setting-sun pattern"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Spoke wheel area",
        "abbreviation": "Misc:SpkWhlA",
        "synonyms": [
            "Globules / Clods : Spoke wheel areas",
            "Lines : Spoke wheel areas",
            "Miscellaneous : Spoke wheel area"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Streaks",
        "abbreviation": "Misc:Streaks",
        "synonyms": [
            "Lines : Streaks",
            "Miscellaneous : Streaks"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : String of pearls",
        "abbreviation": "Misc:SoPearl",
        "synonyms": [
            "Miscellaneous : String of pearls",
            "Vessels : String of pearls"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Structureless, brown (tan)",
        "abbreviation": "Misc:S/lBrwn",
        "synonyms": [
            "Miscellaneous : Structureless, brown (tan)",
            "Structureless : Structureless brown (tan)"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Twisted red loops",
        "abbreviation": "Misc:TwistRL",
        "synonyms": ["Miscellaneous : Twisted red loops"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Miscellaneous : Ulceration / Erosion",
        "abbreviation": "Misc:Ulcer/E",
        "synonyms": ["Miscellaneous : Ulceration / Erosion"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Blood spots",
        "abbreviation": "Nail:BloodSp",
        "synonyms": [
            "Miscellaneous : Blood spots",
            "Nail lesions : Blood spots"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Granular inclusions",
        "abbreviation": "Nail:GranInc",
        "synonyms": ["Nail lesions : Granular inclusions"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Hutchinson sign",
        "abbreviation": "Nail:HutchS.",
        "synonyms": [
            "Miscellaneous : Hutchinson sign",
            "Nail lesions : Hutchinson sign"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Irregular lines",
        "abbreviation": "Nail:Irr.Lin",
        "synonyms": ["Nail lesions : Irregular lines"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Pseudo-Hutchinson sign",
        "abbreviation": "Nail:PseuHut",
        "synonyms": [
            "Miscellaneous : Pseudo-Hutchinson sign",
            "Nail lesions : Pseudo-Hutchinson sign"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Regular lines",
        "abbreviation": "Nail:Reg.Lin",
        "synonyms": ["Nail lesions : Regular lines"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Nail lesions : Splinter hemorrhage",
        "abbreviation": "Nail:SplintH",
        "synonyms": ["Nail lesions : Splinter hemorrhage"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : NOS",
        "abbreviation": "Net:NOS",
        "synonyms": ["Network : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : Atypical pigment network / Reticulation",
        "abbreviation": "Net:AtypPNR",
        "synonyms": [
            "Lines : Atypical pigment network / Reticulation",
            "Network : Atypical pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : Broadened pigment network / Reticulation",
        "abbreviation": "Net:Brd.PNR",
        "synonyms": [
            "Lines : Broadened pigment network / Reticulation",
            "Network : Broadened pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : Delicate Pigment Network / Reticulation",
        "abbreviation": "Net:DeliPNR",
        "synonyms": [
            "Lines : Delicate pigment network / Reticulation",
            "Network : Delicate Pigment Network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : Negative pigment network",
        "abbreviation": "Net:Neg.PN.",
        "synonyms": [
            "Lines : Negative pigment network",
            "Network : Negative pigment network"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Network : Typical pigment network / Reticulation",
        "abbreviation": "Net:Typ.PNR",
        "synonyms": [
            "Lines : Typical pigment network / Reticulation",
            "Network : Typical pigment network / Reticulation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : NOS",
        "abbreviation": "Patt:NOS",
        "synonyms": ["Pattern : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Cerebriform",
        "abbreviation": "Patt:CerebF.",
        "synonyms": [
            "Lines : Cerebriform pattern",
            "Pattern : Cerebriform"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Fibrilar",
        "abbreviation": "Patt:Fibril.",
        "synonyms": [
            "Lines : Fibrillar pattern",
            "Pattern : Fibrilar",
            "Volar lesions : Fibrilar pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Fingerprint",
        "abbreviation": "Patt:FingerP",
        "synonyms": [
            "Lines : Fingerprint pattern",
            "Pattern : Fingerprint"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Homogeneous : NOS",
        "abbreviation": "Patt:Hom.NOS",
        "synonyms": [
            "Pattern : Homogeneous : NOS",
            "Structureless : Homogeneous pattern : NOS"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Homogeneous : Blue",
        "abbreviation": "Patt:Hom.Blu",
        "synonyms": [
            "Pattern : Homogeneous : Blue",
            "Structureless : Homogeneous pattern : Blue"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Homogeneous : Brown",
        "abbreviation": "Patt:Hom.Brw",
        "synonyms": [
            "Pattern : Homogeneous : Brown",
            "Structureless : Homogeneous pattern : Brown"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Homogeneous : Pink",
        "abbreviation": "Patt:Hom.Pnk",
        "synonyms": [
            "Pattern : Homogeneous : Pink",
            "Structureless : Homogeneous pattern : Pink"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Honeycomb",
        "abbreviation": "Patt:Honeycb",
        "synonyms": ["Pattern : Honeycomb"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Latticelike",
        "abbreviation": "Patt:Lattice",
        "synonyms": [
            "Pattern : Latticelike",
            "Volar lesions : Latticelike pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Parallel furrow",
        "abbreviation": "Patt:Par.Fur",
        "synonyms": [
            "Lines : Parallel furrows pattern (volar lesions)",
            "Pattern : Parallel furrow",
            "Volar lesions : Parallel furrow pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Parallel ridge",
        "abbreviation": "Patt:Par.Rid",
        "synonyms": [
            "Lines : Parallel lines (volar lesions)",
            "Pattern : Parallel ridge",
            "Volar lesions : Parallel lines"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Rainbow",
        "abbreviation": "Patt:Rainbow",
        "synonyms": [
            "Pattern : Rainbow",
            "Structureless : Rainbow pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Starburst",
        "abbreviation": "Patt:Starb.",
        "synonyms": [
            "Lines : Starburst pattern",
            "Pattern : Starburst"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Pattern : Strawberry",
        "abbreviation": "Patt:Strawb.",
        "synonyms": [
            "Pattern : Strawberry",
            "Structureless : Strawberry pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Regression structures : NOS",
        "abbreviation": "RegS:NOS",
        "synonyms": ["Regression structures : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Regression structures : Peppering / Granularity",
        "abbreviation": "RegS:PepperG",
        "synonyms": [
            "Dots : Peppering",
            "Regression structures : Peppering / Granularity"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Regression structures : Scarlike depigmentation",
        "abbreviation": "RegS:ScarlkD",
        "synonyms": [
            "Regression structures : Scarlike depigmentation",
            "Structureless : Scar-like depigmentation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Shiny white structures : NOS",
        "abbreviation": "SWhS:NOS",
        "synonyms": ["Shiny white structures : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Shiny white structures : Rosettes",
        "abbreviation": "SWhS:Rosette",
        "synonyms": [
            "Dots : Rosettes",
            "Shiny white structures : Rosettes"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Shiny white structures : Shiny white blotches and strands",
        "abbreviation": "SWhS:ShWhtB.",
        "synonyms": [
            "Globules / Clods : Shiny white blotches and strands",
            "Shiny white structures : Shiny white blotches and strands"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Shiny white structures : Shiny white streaks",
        "abbreviation": "SWhS:ShWhtS.",
        "synonyms": [
            "Lines : Shiny white streaks",
            "Shiny white structures : Shiny white streaks"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : NOS",
        "abbreviation": "SLes:NOS",
        "synonyms": ["Structureless : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Blotch irregular",
        "abbreviation": "SLes:BlotIrr",
        "synonyms": [
            "Miscellaneous : Blotch irregular",
            "Structureless : Blotch irregular"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Blotch regular",
        "abbreviation": "SLes:BlotReg",
        "synonyms": [
            "Miscellaneous : Blotch regular",
            "Structureless : Blotch regular"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Blue-whitish veil",
        "abbreviation": "SLes:Blu-WhV",
        "synonyms": [
            "Miscellaneous : Blue-whitish veil",
            "Structureless : Blue-whitish veil"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Central white patch",
        "abbreviation": "SLes:CentWhP",
        "synonyms": ["Structureless : Central white patch"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Homogeneous pattern : NOS",
        "abbreviation": "SLes:Hom.NOS",
        "synonyms": [
            "Pattern : Homogeneous : NOS",
            "Structureless : Homogeneous pattern : NOS"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Homogeneous pattern : Blue",
        "abbreviation": "SLes:Hom.Blu",
        "synonyms": [
            "Pattern : Homogeneous : Blue",
            "Structureless : Homogeneous pattern : Blue"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Homogeneous pattern : Brown",
        "abbreviation": "SLes:Hom.Brw",
        "synonyms": [
            "Pattern : Homogeneous : Brown",
            "Structureless : Homogeneous pattern : Brown"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Homogeneous pattern : Pink",
        "abbreviation": "SLes:Hom.Pnk",
        "synonyms": [
            "Pattern : Homogeneous : Pink",
            "Structureless : Homogeneous pattern : Pink"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Milky red areas",
        "abbreviation": "SLes:MilkRed",
        "synonyms": ["Structureless : Milky red areas"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Pseudonetwork",
        "abbreviation": "SLes:PseuNet",
        "synonyms": [
            "Facial Skin : Pseudonetwork",
            "Miscellaneous : Pseudonetwork",
            "Structureless : Pseudonetwork"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Rainbow pattern",
        "abbreviation": "SLes:Rainbow",
        "synonyms": [
            "Pattern : Rainbow",
            "Structureless : Rainbow pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Scar-like depigmentation",
        "abbreviation": "SLes:ScarlkD",
        "synonyms": [
            "Regression structures : Scarlike depigmentation",
            "Structureless : Scar-like depigmentation"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Strawberry pattern",
        "abbreviation": "SLes:Strawb.",
        "synonyms": [
            "Pattern : Strawberry",
            "Structureless : Strawberry pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Structureless : Structureless brown (tan)",
        "abbreviation": "SLes:Brown/T",
        "synonyms": [
            "Miscellaneous : Structureless, brown (tan)",
            "Structureless : Structureless brown (tan)"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : NOS",
        "abbreviation": "Vess:NOS",
        "synonyms": ["Vessels : NOS"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Arborizing",
        "abbreviation": "Vess:Arbor.",
        "color": [128,128,32],
        "synonyms": ["Vessels : Arborizing"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Comma",
        "abbreviation": "Vess:Comma",
        "synonyms": ["Vessels : Comma"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Corkscrew",
        "abbreviation": "Vess:Corksc.",
        "synonyms": ["Vessels : Corkscrew"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Crown",
        "abbreviation": "Vess:Crown",
        "synonyms": ["Vessels : Crown"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Dotted",
        "abbreviation": "Vess:Dotted",
        "color": [192,32,32],
        "synonyms": ["Vessels : Dotted"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Glomerular",
        "abbreviation": "Vess:Glomer.",
        "synonyms": ["Vessels : Glomerular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Hairpin",
        "abbreviation": "Vess:Hairpin",
        "synonyms": ["Vessels : Hairpin"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Linear irregular",
        "abbreviation": "Vess:Lin.Irr",
        "synonyms": ["Vessels : Linear irregular"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Monomorphous",
        "abbreviation": "Vess:Monomrp",
        "synonyms": ["Vessels : Monomorphous"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Polymorphous",
        "abbreviation": "Vess:Polymrp",
        "synonyms": ["Vessels : Polymorphous"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : String of pearls",
        "abbreviation": "Vess:SoPearl",
        "synonyms": [
            "Miscellaneous : String of pearls",
            "Vessels : String of pearls"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Vessels : Targetoid",
        "abbreviation": "Vess:Target.",
        "synonyms": ["Vessels : Targetoid"]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Volar lesions : Fibrilar pattern",
        "abbreviation": "VolL:FibrilP",
        "synonyms": [
            "Lines : Fibrillar pattern",
            "Pattern : Fibrilar",
            "Volar lesions : Fibrilar pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Volar lesions : Latticelike pattern",
        "abbreviation": "VolL:Lattice",
        "synonyms": [
            "Pattern : Latticelike",
            "Volar lesions : Latticelike pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Volar lesions : Parallel furrow pattern",
        "abbreviation": "VolL:Par.Fur",
        "synonyms": [
            "Lines : Parallel furrows pattern (volar lesions)",
            "Pattern : Parallel furrow",
            "Volar lesions : Parallel furrow pattern"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Volar lesions : Parallel lines",
        "abbreviation": "VolL:Par.Lin",
        "synonyms": [
            "Lines : Parallel lines (volar lesions)",
            "Pattern : Parallel ridge",
            "Volar lesions : Parallel lines"
        ]
    },
    {
        "nomenclature": "metaphoric",
        "id": "Volar lesions : Parallel ridge pattern",
        "abbreviation": "VolL:Par.Rid",
        "synonyms": [
            "Lines : Parallel ridge pattern (volar lesions)",
            "Volar lesions : Parallel ridge pattern"
        ]
    }
]



"""
isicarchive.segmentation (Segmentation)

This module provides the Segmentation object for the IsicApi to utilize.

Segmentation objects are either returned from calls to

   >>> from isicarchive.api import IsicApi
   >>> api = IsicApi()
   >>> segmentation = api.segmentation(segmentation_id)

or can be generated

   >>> from isicarchive.segmentation import Segmentation
   >>> segmentation = Segmentation(...)
"""

__version__ = '0.4.8'


# imports (needed for majority of functions)
import datetime
import glob
import os
from typing import Any, List, Tuple
import warnings

from . import func
from .vars import ISIC_IMAGE_DISPLAY_SIZE_MAX

_json_full_fields = [
    'created',
    'creator',
    'failed',
    'id',
    'image_id',
    'meta',
    'reviews',
    'skill',
]
_json_partial_fields = [
    'created',
    'id',
    'failed',
    'skill',
]
_mangling = {
    'id': '_id',
    'image_id': 'imageId',
}
_repr_pretty_list = {
    'id': 'id',
    'image_id': 'image_id',
    'image_name': '_image.name',
    'failed': 'failed',
    'meta_source': 'meta.source',
    'reviews': 'reviews.#',
    'reviews_expert_approved': 'reviews.skill=expert.approved',
    'reviews_highest_skill': 'skill',
}
_skill_precedence = {
    'missing': -1,
    'none': 0,
    'novice': 2,
    'expert': 8,
}

class Segmentation(object):
    """
    Segmentation object. If the details are not filled in, only the `id`,
    `created`, `failed`, and `skill` fields will be set.

    To generate a segmentation object for an existing dataset, please use
    the IsicApi.segmentation(...) method!

    To generate a new segmentation object (for later storage), use

       >>> segmentation = Segmentation(image_id=image_id, data=...)

    Attributes
    ----------
    created : Date
        Study creation date (w.r.t. in the database!)
    creator : dict
        Contains _id and (short) name field identifying the creator
    failed : bool
        Flag whether segmentation (approval?) failed
    id : str
        mongodb objectId of the segmentation
    image_id : str
        mongodb objectId of the associated image
    mask : numpy.ndarray (or None)
        Segmentation mask data
    meta : dict
        Metadata associated with the segmentation, containing fields
        'source' - token describing the source of the segmentation,
                   e.g. 'autofill'
        'startTime' - when the segmentation began
        'stopTime - when the segmentation was completed
    reviews : list
        List of dicts with fields 'approved', 'skill', 'time', and 'userId'
    
    Methods
    -------
    """


    def __init__(self,
        from_json:dict = None,
        name:str = None,
        api:object = None,
        load_mask_data:bool = False,
        ):
        """Segmentation init."""

        self._api = api
        self._detail = False
        self._image = None
        self._image_obj = None
        self._in_archive = False
        self._model_type = 'segmentation'
        self._raw_data = None
        self._sp_in_mask = None
        # still needs timezone information!!
        self.area = 0
        self.area_pct = 0.0
        self.created = datetime.datetime.now().strftime(
            '%Y-%m-%dT%H:%M:%S.%f+00:00')
        self.creator = {'_id': '000000000000000000000000'}
        self.failed = True
        self.id = ''
        self.image_id = ''
        self.mask = None
        self.meta = dict()
        self.reviews = []
        self.skill = 'none'

        # from JSON
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json)
            except:
                raise
        if self._in_archive and load_mask_data:
            try:
                self.load_mask_data()
            except Exception as e:
                warnings.warn(str(e))

    # parse JSON
    def _from_json(self, from_json:dict):
        self.id = from_json['_id']
        if 'created' in from_json:
            self.created = from_json['created']
        if 'creator' in from_json:
            self.creator = from_json['creator']
        if 'failed' in from_json:
            self.failed = from_json['failed']
        if 'imageId' in from_json:
            self.image_id = from_json['imageId']
            if self.image_id in self._api._image_objs:
                self._image_obj = self._api._image_objs[self.image_id]
            if self.image_id in self._api.image_cache:
                self._image = self._api.image_cache[self.image_id]
        if 'meta' in from_json:
            self.meta = from_json['meta']
        if 'reviews' in from_json:
            self.reviews = from_json['reviews']
            self._detail = True
            max_skill = 0
            for review in self.reviews:
                if ('approved' in review and review['approved']
                    and review['skill'] in _skill_precedence):
                    if _skill_precedence[review['skill']] > max_skill:
                        self.skill = review['skill']
                        max_skill = _skill_precedence[self.skill]
        if 'skill' in from_json:
            self.skill = from_json['skill']
        self._in_archive = True

    # JSON
    def __repr__(self):
        return 'isicarchive.segmentation.Segmentation(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Segmentation (id={0:s}, image_id={1:s})'.format(
            self.id, self.image_id)
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)
    
    # JSON representation (without constructor):
    def as_json(self):

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        from json import dumps as json_dumps

        json_list = []
        fields = _json_full_fields if self._detail else _json_partial_fields
        for field in fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json_dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'

    # clear data
    def clear_data(self,
        clear_raw_data:bool = True,
        clear_mask:bool = True,
        clear_superpixels_info:bool = True,
        deref_image:bool = False):
        if deref_image:
            self._image = None
            self._image_obj = None
        if clear_raw_data:
            self._raw_data = None
        if clear_mask:
            self.mask = None
        if clear_superpixels_info:
            self._sp_in_mask = None

    # load mask data
    def load_mask_data(self, keep_raw_data:bool = False):

        # IMPORTS DONE HERE TO SAVE TIME AT MODULE INIT
        import imageio
        import numpy

        if not self._api:
            raise ValueError('Invalid segmentation object to load mask data for.')
        if self._api._cache_folder:
            smask_filename = self._api.cache_filename(self.id, 'smask', '.*', '*')
            smask_list = glob.glob(smask_filename)
            if smask_list:
                try:
                    self.data = None
                    with open(smask_list[0], 'rb') as mask_file:
                        mask_raw = mask_file.read()
                    if keep_raw_data:
                        self._raw_data = mask_raw
                    self.mask = imageio.imread(mask_raw)
                    self.area = numpy.sum(self.mask > 0)
                    self.area_pct = self.area / self.mask.size
                    return
                except Exception as e:
                    warnings.warn('Error loading segmentation mask: ' + str(e))
                    os.remove(smask_list[0])
        if self._in_archive and self._api:
            try:
                req = self._api.get('segmentation/' + self.id + '/mask',
                    parse_json=False)
                if req.ok:
                    mask_raw = req.content
                    if keep_raw_data:
                        self._raw_data = mask_raw
                    self.mask = imageio.imread(mask_raw)
                    self.area = numpy.sum(self.mask > 0)
                    self.area_pct = self.area / self.mask.size
                    if self._api._cache_folder:
                        if not self._image_obj is None and (len(self._image_obj.name) > 5):
                            extra = self._image_obj.name
                        elif not self._image is None and (len(self._image['name']) > 5):
                            extra = self._image['name']
                        else:
                            extra = None
                        mask_filename = self._api.cache_filename(self.id,
                            'smask', func.guess_file_extension(req.headers),
                            extra)
                        with open(mask_filename, 'wb') as mask_file:
                            mask_file.write(mask_raw)
            except Exception as e:
                warnings.warn('Error loading segmentation mask: ' + str(e))

    # segmentation outline
    def outline(self,
        out_format:str = 'osvg',
        negative:bool = True,
        ) -> Any:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        from .imfunc import segmentation_outline

        if self.mask is None:
            try:
                self.load_mask_data()
                if self.mask is None:
                    raise RuntimeError('Error loading mask.')
            except:
                raise
        return segmentation_outline(self.mask,
            out_format=out_format, negative=negative)

    # show image in notebook
    def show_in_notebook(self,
        on_image:bool = True,
        mask_color:Tuple = None,
        max_size:int = None,
        library:str = 'matplotlib',
        call_display:bool = True,
        ) -> object:

        # IMPORTS DONE HERE TO SAVE TIME AT MODULE INIT
        import numpy
        from .imfunc import display_image, image_mix
        try:
            if self.mask is None:
                self.load_mask_data()
            image_data = self.mask
            mask = (image_data == 0)
            if isinstance(mask_color, tuple) and len(mask_color) == 3:
                shp = image_data.shape
                image_data = image_data.reshape(
                    (shp[0] * shp[1], 1,)).repeat(3, axis=1)
                image_data[mask.reshape(shp[0] * shp[1], 1), :] = [
                    mask_color[0], mask_color[1], mask_color[2]]
                image_data.shape = (shp[0], shp[1], 3)
            if on_image:
                if not self._image_obj:
                    if self.image_id in self._api._image_objs:
                        self._image_obj = self._api._image_objs[self.image_id]
                    else:
                        self._image_obj = self._api.image(self.image_id)
                self._image_obj.load_image_data()
                alpha = numpy.zeros(image_data.shape, dtype=numpy.float32, order='C')
                alpha[image_data == 0] = 0.5
                image_data = image_mix(self._image_obj.data, image_data, alpha)
            return display_image(image_data, max_size=max_size,
                ipython_as_object=(not call_display), library=library)
        except Exception as e:
            warnings.warn('show_in_notebook(...) failed: ' + str(e))

    # superpixels in mask
    def superpixels_in_mask(self, thresh:float=None) -> List:

        # IMPORT DONE HERE TO SAVE TIME AT MODULE INIT
        import numpy

        if self._sp_in_mask is None:
            if self.mask is None:
                self.load_mask_data()
            mask = self.mask.reshape((self.mask.size,))
            if self._image_obj is None:
                try:
                    self._image_obj = self._api.image(self.image_id)
                except:
                    warnings.warn('Unable to retrieve image object')
                    return
            if self._image_obj.superpixels['map'] is None:
                self._image_obj.map_superpixels()
            sp_map = self._image_obj.superpixels['map']
            sp_list = [0.0] * sp_map.shape[0]
            for sp in range(len(sp_list)):
                mask_val = (mask[sp_map[sp,0:sp_map[sp,-1]]] > 0)
                sp_list[sp] = float(numpy.sum(mask_val)) / float(mask_val.size)
            self._sp_in_mask = sp_list
        if thresh is None:
            return self._sp_in_mask[:]
        return [idx for (idx, val) in enumerate(self._sp_in_mask) if val >= thresh]
